/* FG - Flaechen-Generator  */

/*
**
**  INCLUDE FILES
**
*/

#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include "tmdef.h"
#include "otdef.h"


/*
**
**  MACRO DEFINITIONS
**
*/

#define repeat	for(;;)
#define assgk(p1, p2)	{short I;for (I=0;I<DIM;I++) p1[I]=p2[I];}
#define MAXBUF 250		    /* max. Pufferlaenge fuer Formel */
#define MAXTMP	50		    /* max. Anzahl von temporaeren Strings fuer Formel definition */
#define MAXNPF  150		    /* max. Punktzahl einer Facette */
#define GRAD	0.0174532925199	    /* pi/180 */

extern OTwurzel otw;		    /* Quad-tree-Wurzel */

extern int N_Fcall;         /* Anzahl der Funktionsaufrufe */
int        N_Wref;          /* Anzahl der Wert-Referenzen  */

extern char *getenv();

FILE *fpp;

TMtab *ktab,			    /* Koordinatentabelle */
      *ptab;			    /* Schnittpunkttabelle */

OTpunkt epsd1, 
	epsd2,
	epsd3 = 0.0005,		    /* Schranke fuer Iterationsgenauigkeit */
	epsd4 = 0.0001,		    /* Schranke fuer Punktidentifikation   */
	epsdr,
	epsdc,
	epsdm = GRAD,		    /* Kollinearitaet bei Winkel <= 1 Grad */
	epsdp = 0.05234,	    /* Planaritaetsabweichung < 3 Grad */
	epsdd = 0.00137;	    /* Divergenz < 3 Grad */

short int kmax1, 
	  kmax2,
	  kmaxr,
	  pchk = 0, dchk = 0, cchk = 0, schk = 0, tchk = 0, tcheck = 0,
	  trans = 0,
      w_in_grad = 0,
	  nv_mod,
	  a_mod = 1;
                                   /* a_mod :  1-> Kondar (nicht benutzt)
                                               2-> PPatch (max Vierecke)
                                               3-> Patch  (n-Ecke)
                                               4-> Povray (smooth triangles)
                                   */ 

short int Iter = 0,
	  Iter1 = 8,
	  Iter2 = 8;
int Nzkomp = 0;			    /* Anzahl der Zusammenhangskomponenten */

double	Rot[3][3], pv[3], ps[3];

jmp_buf restart;		    /* Neustart-Adresse bei Syntax-Fehler */

#ifdef vms
typedef struct
{
   int len;
   char *txt;
} String;
#else
typedef char String;
#endif

extern double atof(),
	      Vskal(),
	      F_val();

#ifdef KONDAR
extern short int KMSaus(),
		 KMSdfc();

extern void KMSext();
#endif

short int FG_init(),
	  FG_gfac();

#ifdef vms
extern void lib$init_timer(), lib$show_timer();
#endif

main ()
{
   short int i, k, L,
	     eastat,
	     npol;
   int n1, n2, n3, n4;
   unsigned int nfp;
   OTpunkt Orig[DIM], 
	   sl, 
	   d1, d2, dr, h1, h2, h3,
	   mxz, mnz;
   double  vr[4];
   char txt1[MAXBUF], 
	txt2[100], 
	*txtf, 
	*tp,
	*tmptxt[MAXTMP];
   String name;
   float bf;

   printf("** FG - Flaechen-Generator\n++ Vordefinierte Variablen:\n");
   txtf = NULL;  sl = .0;

   if ( F_ini("?s#"))
      exit();

   if (tp = getenv("FG_epsdm"))
      epsdm = atof(tp) * GRAD;

   if (tp = getenv("FG_epsdi"))
      epsd4 = atof(tp);

   printf("++ Schranke fuer Punktidentifikation: %g\n", epsd4 );
   printf("++ Schranke fuer Kollinearitaet: %g\n", epsdm );

   setjmp(restart);
   printf("\n");

   do
   {
      repeat
      {
	 printf("-- Formel-Initialisierung\n");
	    /* Eine oder mehrere Kalkulatoranweisungen, abgeschlossen
	       mit einer Leerzeile */

	 if ((tp = fgets(txt1,MAXBUF,stdin)) == NULL || strlen(tp) < 3)
	    break;

	 txt1[strlen(txt1)-1] = '#';

	 if ( F_ini(txt1))
	    exit();
      }

      printf("-- Formel\n");
	 /* Auszuwertende Formel - kann ueber mehrere Zeilen fortgesetzt
	    werden (max. 50 Zeilen), abgeschlossen
	    mit einer Leerzeile */

      if ((tp = fgets(txt1,MAXBUF,stdin)) && (k = strlen(tp)) > 2)
      {
	 if (txtf)
	    free(txtf);

	 txtf = NULL;
	 i = L = 0;

	 repeat
	 {
	    txt1[--k] = '\0';		/* Linefeed loeschen */
	    L += k;
	    tmptxt[i] = emalloc((unsigned) k+1);
	    strcpy( tmptxt[i++], txt1 );

	    if ( !(tp = fgets(txt1,MAXBUF,stdin)) || (k = strlen(tp)) < 3)
	       break;
	 }

	 if ( L < 3 )
	 {
	    printf("!! Keine Formel definiert !\n" );
	    continue;
	 }

	 txtf = emalloc((unsigned) L+2);
	 *txtf = '\0';

	 for (k = 0;  k < i;  k++)
	 {
	    strcat( txtf, tmptxt[k]);
	    free(tmptxt[k]);
	 }

	 txtf[L] = '#';   txtf[L+1] = '\0';
      }
      else if ( txtf == NULL )
      {
	 printf("!! Keine Formel definiert !\n" );
	 continue;
      }

      printf("-- Orig, sl, d1, d2\n");
	 /* Orig:    Koordinaten des Wuerfelursprungs,
	    sl:	     Seitenlaenge des Wuerfels,
	    d1:	     max. Teilwuerfelgroesse fuer Detailerkennung,
	    d2:	     max. Teilwuerfelgroesse insgesamt
	 */

      if ((tp = fgets(txt2,100,stdin)) && strlen(tp) > 1)
      {
	 Orig[0] = atof(tp);    tp = strchr(tp, ',') + 1;
	 Orig[1] = atof(tp);    tp = strchr(tp, ',') + 1;
	 Orig[2] = atof(tp);    tp = strchr(tp, ',') + 1;
	 sl    = atof(tp);    tp = strchr(tp, ',') + 1;
	 d1    = atof(tp);    tp = strchr(tp, ',') + 1;
	 d2    = atof(tp);
      }

      if ( sl < .001 )
      {
	 printf("!! Kein gueltiger Arbeitsbereich definiert !\n" );
	 continue;
      }

      printf("-- epsp, epsd, epsc, Schk, dr, Tchk, Ausg, Trans\n");
	 /* epsp:    max. Winkelabweichung (in Grad) fuer Planaritaetstest,
	    epsd:    max. Winkelabweichung (in Grad) fuer Divergenztest,
	    epsc:    max. chordaler Abstand,
	    Schk:    0/1 = ohne/mit Singularitaetstests,
	    dr:	     max. Teilwuerfelgroesse fuer Reduktion,
	    Tchk:    0/1 = ohne/mit Topologietest
	    Ausg:    1 = Kondar-Modell, 2 = Ppatch-Datei
                     3 = Patch-Datei,   4 = Povray-Datei
	    Trans:   0 = ohne Transformation, 1 = mit Dilatation,
		     2 = mit Rotation, 4 = mit Translation
	 */

      trans = 0;

      if ((tp = fgets(txt2,100,stdin)) && strlen(tp) > 1)
      {
	 h1    = atof(tp);    tp = strchr(tp, ',') + 1;
	 h2    = atof(tp);    tp = strchr(tp, ',') + 1;
	 h3    = atof(tp);

	 if (strlen(tp) > 1 && (tp = strchr(tp, ',')))
	 {
	    schk = ( atoi(++tp) != 0 );

	    if (strlen(tp) > 1 && (tp = strchr(tp, ',')))
	    {
	       dr = atof(++tp);

	       if (strlen(tp) > 1 && (tp = strchr(tp, ',')))
	       {
		  tchk = ( atoi(++tp) != 0 );

		  if (strlen(tp) > 1 && (tp = strchr(tp, ',')))
		  {
		     a_mod = atoi(++tp);

		     if (strlen(tp) > 1 && (tp = strchr(tp, ',')))
			trans = atoi(++tp);
		  }
		  else
		     a_mod = 1;
	       }
	       else
	       {
		  tchk = 0;
		  a_mod = 1;
	       }
	    }
	    else
	    {
	       dr = 0.;
	       tchk = 0;
	       a_mod = 1;
	    }
	 }
	 else
	 {
	    tchk = schk = 0;
	    dr = d1;
	    a_mod = 1;
	 }

	 pchk  = (h1 > 0.0);  dchk = (h2 > 0.0);  cchk = (h3 > 0.0);

	 if (pchk)
	    epsdp = cos((90-h1)*GRAD);
	 if (dchk)
	    epsdd = 1.0 - cos(h2*GRAD);

	 if (cchk)
	    epsdc = h3;

	 if (trans)
	 {
	    if (trans & 1)
	    {
	       printf("-- sx, sy, sz\n");

	       if ((tp = fgets(txt2,100,stdin)) && strlen(tp) > 1)
	       {
		  ps[0] = atof(tp);    tp = strchr(tp, ',') + 1;
		  ps[1] = atof(tp);    tp = strchr(tp, ',') + 1;
		  ps[2] = atof(tp);
	       }
	    }

	    if (trans & 2)
	    {
	       printf("-- u, v, w\n");

	       if ((tp = fgets(txt2,100,stdin)) && strlen(tp) > 1)
	       {
		  h1    = atof(tp);    tp = strchr(tp, ',') + 1;
		  h2    = atof(tp);    tp = strchr(tp, ',') + 1;
		  h3    = atof(tp);
		  h3 *= (0.5*GRAD);
		  vr[0] = cos(h3);
		  Vkuka(h1*GRAD, h2*GRAD, sin(h3), &vr[1]);
		  Qqnm(vr,Rot);

		  for ( i = 0; i < 3; i++ )
		     for ( k = 0; k < 3; k++ )
			if ( fabs(Rot[i][k]) < 1e-6 )
			   Rot[i][k] = 0.0;
	       }
	    }

	    if (trans & 4)
	    {
	       printf("-- vx, vy, vz\n");

	       if ((tp = fgets(txt2,100,stdin)) && strlen(tp) > 1)
	       {
		  pv[0] = atof(tp);    tp = strchr(tp, ',') + 1;
		  pv[1] = atof(tp);    tp = strchr(tp, ',') + 1;
		  pv[2] = atof(tp);
	       }
	    }
	 }
      }

      F_def(txtf, "x,y,z", .0);

      if (FG_init(Orig, sl, d1, d2, dr) < 0)
      {
	 printf("!! Speichermangel -> Abbruch\n");
	 exit();
      }

      printf("++ 1. Teilung: %d, eps = %g\n   2. Teilung: %d, eps = %g\n   3. Teilung: %d, eps = %g\n",
		 kmax1, epsd1, kmax2, epsd2, kmaxr, epsdr);
      printf("-- Mit Nachiteration (j/n) ?:\n");

      if (Iter = ((tp = fgets(txt2,10,stdin)) && strlen(tp) > 1 && *tp == 'j'))
      {
	 if (strlen(++tp) > 1)
	    Iter1 = atoi(tp);

	 if ((tp = strchr(tp, ',')) && strlen(++tp) > 1 )
	    Iter2 = atoi(tp);
      }
      else
	 Iter1 = Iter2 = 0;

#ifdef vms
      lib$init_timer();
      n4 = N_Fcall;
#endif
      tcheck = 0;
      nv_mod = ( dchk || (a_mod >= 2) );
      OTS_part(otw.start);		 /* Octree erzeugen */
      n1 = n2 = 0;
      OTA_nel(otw.start, &n1, &n2);
      n3 = (n1 * sizeof(OTknoten))/1024;
      printf("++ Zerlegungsbaum erzeugt.\n   Anzahl der Knoten: %d (%d) - %d Kb\n", n1, n2, n3 );
#ifdef vms
      printf("++ Zeitbedarf (incl. %d Funktionsaufrufen):\n", N_Fcall-n4);
      lib$show_timer();
#endif
/*
      FG_Ncheck();
*/
      if (tchk)
      {
	 tcheck = 1;
	 OTS_tcheck(otw.start);		  /* Octree auf topologische Kriterien testen*/
	 n3 = n4 = 0;
	 OTA_nel(otw.start, &n3, &n4);

	 if (n3 > n1)
	 {
	    n2 = (n3 * sizeof(OTknoten))/1024;
	    printf("++ Zerlegungsbaum erweitert.\n   Anzahl der Knoten: %d (%d) - %d Kb\n", n3, n4, n2 );
	    n1 = n3;
	 }
	 else
	   printf("++ Topologietest erfolglos\n");
      }

      if (kmaxr)
      {
	 OTS_join(otw.start);		  /* Octree reduzieren */
	 n3 = n4 = 0;
	 OTA_nel(otw.start, &n3, &n4);

	 if (n3 < n1)
	 {
	    n2 = (n3 * sizeof(OTknoten))/1024;
	    printf("++ Zerlegungsbaum reduziert.\n   Anzahl der Knoten: %d (%d) - %d Kb\n", n3, n4, n2 );
	 }
	 else
	   printf("++ Reduktion erfolglos\n");
      }

#ifdef KONDAR
      if (a_mod == 1)
	 KMSini();
#endif

      if (a_mod >= 2)
	 fpp = fopen("facetten.tmp", "w");

#ifdef vms
      lib$init_timer();
      n4 = N_Fcall;
#endif
      nv_mod = (a_mod >= 2);
      k = FG_gfac(&nfp);

      if ((a_mod >= 2) && fpp)
	 fclose(fpp);

      if (a_mod == 1)
	 printf("++ %d Facetten erzeugt.\n", k);

      if (a_mod >= 2)
	 printf("++ %d Patche erzeugt.\n", nfp);

      printf("++ %d Funktionsaufrufe insgesamt\n", N_Fcall);
#ifdef vms
      printf("   Zeitbedarf (incl. %d Funktionsaufrufen):\n", N_Fcall-n4);
      lib$show_timer();
#endif
      n1 = n2 = 0;
      TMA_stat(ktab, &n1, &N_Wref, &n2, &n3, &bf, &mxz, &mnz);
      printf("++ Extremale Funktionswerte: %g, %g\n", mxz, mnz);
      printf("++ Anzahl berechneter Gitterpunkte: %d - %d Kb\n", n1, (n1 * sizeof(TMelem))/1024);
      printf("++ Tabellenwerte (Groesse, Belegungsfaktor, Konflikte, max. Kettenlaenge):\n   %d, %f, %d, %d\n", 
		 ktab->tsize, bf, n3, n2+1);
      n1 = n2 = 0;
      TMA_stat(ptab, &n1, &n4, &n2, &n3, &bf, &mxz, &mnz);
      printf("++ Anzahl berechneter Schnittpunkte: %d - %d Kb\n", n1, (n1 * sizeof(TMelpk))/1024);
      printf("++ Tabellenwerte (Groesse, Belegungsfaktor, Konflikte, max. Kettenlaenge):\n   %d, %f, %d, %d\n", 
		 ptab->tsize, bf, n3, n2+1);
      printf("++ %d Zugriffe auf Funktionswerte\n   %d Zugriffe auf Schnittpunkte\n",
		    N_Wref, n4);
      printf("   %d Zusammenhangskomponente(n)\n-- Name der Flaeche:\n", Nzkomp);
#ifdef vms
      name.len = 20;
      name.txt = txt2;
#else
#define name txt2
#endif

      if ((tp = fgets(txt2,100,stdin)) && (L = strlen(tp)-1) > 0)
      {
	 txt2[L] = '#';
#ifndef MOD_TETRA
#ifdef KONDAR
	 if (a_mod == 1)
	 {
	    KMSext();

	    if (k = KMSaus(&name, &eastat))
	       printf("!! Fehler bei Modellausgabe: %d\n", eastat );
	 }
#endif

	 if (a_mod >= 2)
	 {
	    txt2[L] = '\0';
/*	    if (a_mod == 2) strcat( txt2, ".ppa" );*/
	    rename( "facetten.tmp", txt2 );
	 }
#endif
      }

      printf("-- Weitere Berechnung (j/n) ?:\n");

      if (k = ((tp = fgets(txt2,10,stdin)) && strlen(tp) > 1 && *tp == 'j'))
      {
	 OTS_del(otw.start);
	 TMS_del(ktab);
	 TMS_del(ptab);
      }
   }
   while ( k );
}

static short int				 /* Rueckmeldung: 0 = ok, -1 = Fehler */
FG_init (Orig, sl, d1, d2, dr)			 /* Initialisiere Flaechen-Generator */
OTpunkt Orig[],					 /* (E) Linke untere Ecke */
	sl,					 /* (E) Seitenlaenge */
	d1,					 /* (E) kleinstes zu erfassendes Detail */
	d2,					 /* (E) max. Aufloesung der Flaeche(n) */
	dr;					 /* (E) max. Aufloesung regulaerer Gebiete */
{
   short int k;
   TMkey i1, i2, i3,
	 key[DIM], 
	 smax[DIM], 
	 d3, dd;
   TMelem *q;

   kmax1 = OTA_maxT(sl, d1, &epsd1);		 /* Gitterkennwerte */
   kmax2 = OTA_maxT(sl, d2, &epsd2);
   kmaxr = (dr > 0.0 ? OTA_maxT(sl, dr, &epsdr) : 0);
   d3 = (1 << kmax2);
   dd = (1 << ((kmax2+kmax1)/2));
   smax[0] = smax[1] = smax[2] = dd;

   if ((ktab = TMS_def(3, smax)) == NULL)			 /* Koordinaten-Tabelle definieren */
      return(-1);

   i1 = dd >> 2;
   smax[0] = smax[1] = smax[2] = ( i1 > 12 ? i1 : 12 );

   if ((ptab = TMS_def(6, smax)) == NULL)			 /* Schnittpunkt-Tabelle definieren */
      return(-1);

   OTS_ini(Orig, sl, epsd2, kmax1, kmax2, kmaxr);
   otw.start = OTS_neu();
   otw.start->level = 0;
   otw.start->Orig[0] = otw.start->Orig[1] = otw.start->Orig[2] = 0;
   dd = otw.kd1max;
#ifdef vms
   lib$init_timer();
#endif

   if (OTS_Tixe())
      for (i1 = 0;  i1 <= d3;  i1 += dd)
      {
	 key[0] = i1;

	 for (i2 = 0;  i2 <= d3;  i2 += dd)
	 {
	    key[1] = i2;

	    for (i3 = 0;  i3 <= d3;  i3 += dd)
	    {
	       key[2] = i3;

	       if ((q = TMS_ins(ktab, key, &k)) && k == 0 )
		  OTS_gkv(q);
	       else
		  return ( -1 );
	    }
	 }
      }

#ifdef vms
   printf("++ Start-Gitter berechnet (%d Funktionsaufrufe). Zeitbedarf:\n", N_Fcall);
   lib$show_timer();
#endif

   return ( 0 );
}

PRIVATE short FG_fac(np,polin,rpol,polout,nv)	  /* Facette umspeichern und nachbearbeiten */
short np, rpol[];
OTpol  polin[];
float  polout[][DIM], nv[][DIM];
{
   short int i, j, k, nnp;
   double   vs, vp[3], v1[3], v2[3];
   TMelpk *e;

   for ( nnp = i = 0; i < np; i++ )    /* 1. Transformation und Redundanz-Reduktion */
   {
      if (trans)
      {
	 if (trans & 1)
	    for ( k = 0; k < DIM; k++ )
	       polin[i].sp[k] *= ps[k];

	 if (trans & 2)
	 {
	    Vmmul( Rot, polin[i].sp, vp );

	    for ( k = 0; k < DIM; k++ )
	       polin[i].sp[k] = vp[k];
	 }

	 if (trans & 4)
	    for ( k = 0; k < DIM; k++ )
	       polin[i].sp[k] += pv[k];
      }

      if (i && epsd4 > 0.0)
      {
	 for ( j = 1, k = 0; j && k < DIM; k++ )
	    j = ( fabs(polin[i].sp[k] - polin[rpol[nnp-1]].sp[k]) <= epsd4 );

	 if (!j)
	    rpol[nnp++] = i;
      }
      else
	 rpol[nnp++] = i;
   }

   if ( nnp > 3 )
   {
      np = nnp;
      Vdif( polin[rpol[1]].sp, polin[rpol[0]].sp, v1);
      Vricht(v1,v1);

      for ( nnp = 1, i = 2; i < np; i++ )    /* 2. Kollinearitaetstest */
      {
	 Vdif( polin[rpol[i]].sp, polin[rpol[i-1]].sp, v2);
	 Vricht(v2,v2);
	 Vprod( v1, v2, vp );

	 if ((vs = Vskal(vp,vp)) > epsdm )
	 {
	    if ( i - 1 > nnp )
	       rpol[nnp] = rpol[i - 1];

	    nnp++;
	    assgk(v1,v2);
	 }
      }

      if ( np - 1 > nnp )
	 rpol[nnp] = rpol[np - 1];

      nnp++;
   }

   if (nnp > 2)
      for ( i = 0; i < nnp; i++ )
      {
	 assgk(polout[i], polin[rpol[i]].sp);

	 if (a_mod >= 2)
	 {
	    if ((e = TMA_ele(ptab, polin[rpol[i]].sk, &k)) && k == 1 && e->nv )
	    {
	       if (trans & 2)
	       {
		  Vmmul( Rot, e->nv, vp);
		  assgk(nv[i], vp);
	       }
	       else
	       {
		  assgk(nv[i], e->nv);
	       }
	    }
	    else
	       printf( "!! Fehler (FG_fac): Normalvektor nicht gefunden!\n");
	 }
      }

   return ( nnp );
}

PRIVATE void
Form_konv( zahl, text )
float zahl;
char  text[];
{
   short i;

   sprintf( text, "%.5f", (fabs(zahl) < 0.0001 ? 0.0 : zahl));

   for ( i = strlen(text); i > 2 && text[i-1] == '0'; --i )
      text[i-1] = '\0';

   if ( text[i-1] == '.' )
      text[i-1] = '\0';

   if ( text[0] == '0' && text[1] == '.' )
      for ( i = 1; text[i-1] = text[i]; i++ )
	 ;
   else if ( text[0] == '-' && text[1] == '0' && text[2] == '.' )
      for ( i = 2; text[i-1] = text[i]; i++ )
	 ;

   if ( text[0] == '-' && text[1] == '0' && text[2] == '\0' )
      for ( i = 1; text[i-1] = text[i]; i++ )
	 ;
}

PRIVATE void
FG_ppatch(np,fac,nv,nfacp)	  /* Ppatch erzeugen */
short np;
float  fac[][DIM], nv[][DIM];
unsigned int *nfacp;
{				 /* Achtung: fuer np > 4 werden fac und nv veraendert ! */
   short int i, j, k, kk;
   char  txt[20];
   double vs, vp[3], ww[4], vv[4][3];

   if (np <= 4)
   {
      for (j = 0;  j < DIM;  j++)
      {
	 vv[0][j] = fac[1][j] - fac[0][j];
	 vv[1][j] = fac[2][j] - fac[1][j];
      }

      Vricht(vv[0], vv[0]);
      Vricht(vv[1], vv[1]);
      Vprod(vv[0], vv[1], vp);

      if ((ww[0] = Vskal(vp, vp)) < 5 * GRAD)
      {
	 for (i = 2;  i < np;  i++)
	 {
	    k = ( i+1 < np ? i+1 : 0 );

	    for (j = 0;  j < DIM;  j++)
	       vv[i][j] = fac[k][j] - fac[i][j];

	    Vricht(vv[i], vv[i]);
	 }

	 for (kk = 0, i = 1;  i < np;  i++)
	 {
	    k = ( i+1 < np ? i+1 : 0 );
	    Vprod(vv[i], vv[k], vp);
	    ww[i] = Vskal(vp, vp);

	    if (ww[i] > ww[kk])
	       kk = i;
	 }

	 k = kk;
      }
      else
	 k = 0;

      fprintf( fpp, "ppatch %d\n", np);

      for (i = 0;  i < np;  i++)
      {
	 for (j = 0;  j < DIM;  j++)
	 {
	    Form_konv( fac[k][j], txt);
	    fprintf( fpp, "%s%s", (j ? " " : ""), txt);
	 }

	 for (j = 0;  j < DIM;  j++)
	 {
	    Form_konv( nv[k][j], txt);
	    fprintf( fpp, " %s", txt);
	 }

	 fprintf( fpp, "\n");
	 k = (k+1 < np ? k+1 : 0);
      }

      (*nfacp)++;
   }
   else
   {
      FG_ppatch(4, fac, nv, nfacp);

      for (j = 0;  j < DIM;  j++)
      {
	 fac[2][j] = fac[0][j];
	 nv[2][j] = nv[0][j];
      }

      FG_ppatch(np - 2, &fac[2], &nv[2], nfacp);
   }
}

PRIVATE void
FG_patch(np,fac,nv,nfacp)	  /* Ppatch erzeugen */
short np;
float  fac[][DIM], nv[][DIM];
unsigned int *nfacp;
{				 /* Achtung: fuer np > 4 werden fac und nv veraendert ! */
   short int i, j, k, kk;
   char  txt[20];
   double vs, vp[3], ww[6], vv[6][3];

/********* auch Patches mit mehr als vier Punkten ausgeben **********/

   if (np <= 6)   /* also immer (hoffentlich) */
   {
      for (j = 0;  j < DIM;  j++)
      {
	 vv[0][j] = fac[1][j] - fac[0][j];
	 vv[1][j] = fac[2][j] - fac[1][j];
      }

      Vricht(vv[0], vv[0]);
      Vricht(vv[1], vv[1]);
      Vprod(vv[0], vv[1], vp);

      if ((ww[0] = Vskal(vp, vp)) < 5 * GRAD)
      {
	 for (i = 2;  i < np;  i++)
	 {
	    k = ( i+1 < np ? i+1 : 0 );

	    for (j = 0;  j < DIM;  j++)
	       vv[i][j] = fac[k][j] - fac[i][j];

	    Vricht(vv[i], vv[i]);
	 }

	 for (kk = 0, i = 1;  i < np;  i++)
	 {
	    k = ( i+1 < np ? i+1 : 0 );
	    Vprod(vv[i], vv[k], vp);
	    ww[i] = Vskal(vp, vp);

	    if (ww[i] > ww[kk])
	       kk = i;
	 }

	 k = kk;
      }
      else
	 k = 0;

      fprintf( fpp, "ppatch %d\n", np);

      for (i = 0;  i < np;  i++)
      {
	 for (j = 0;  j < DIM;  j++)
	 {
	    Form_konv( fac[k][j], txt);
	    fprintf( fpp, "%s%s", (j ? " " : ""), txt);
	 }

	 for (j = 0;  j < DIM;  j++)
	 {
	    Form_konv( nv[k][j], txt);
	    fprintf( fpp, " %s", txt);
	 }

	 fprintf( fpp, "\n");
	 k = (k+1 < np ? k+1 : 0);
      }

      (*nfacp)++;
   }
   else
   {
      FG_patch(4, fac, nv, nfacp);

      for (j = 0;  j < DIM;  j++)
      {
	 fac[2][j] = fac[0][j];
	 nv[2][j] = nv[0][j];
      }

      FG_patch(np - 2, &fac[2], &nv[2], nfacp);
   }
}


PRIVATE void
FG_povray_triangle(np,fac,nv,nfacp)	  /* Ppatch erzeugen */
short np;
float  fac[][DIM], nv[][DIM];
unsigned int *nfacp;
{				 /* Achtung: fuer np > 4 werden fac und nv veraendert ! */
   short int i, j, k, kk,jj;
   char  txt[20];
   double vs, vp[3], ww[20], vv[20][3];

/********* nur Dreiecke ausgeben **********/

   if (np <= 3)   
   {
      for (j = 0;  j < DIM;  j++)
      {
	 vv[0][j] = fac[1][j] - fac[0][j];
	 vv[1][j] = fac[2][j] - fac[1][j];
      }

      Vricht(vv[0], vv[0]);
      Vricht(vv[1], vv[1]);
      Vprod(vv[0], vv[1], vp);

      if ((ww[0] = Vskal(vp, vp)) < 5 * GRAD)
      {
	 for (i = 2;  i < np;  i++)
	 {
	    k = ( i+1 < np ? i+1 : 0 );

	    for (j = 0;  j < DIM;  j++)
	       vv[i][j] = fac[k][j] - fac[i][j];

	    Vricht(vv[i], vv[i]);
	 }

	 for (kk = 0, i = 1;  i < np;  i++)
	 {
	    k = ( i+1 < np ? i+1 : 0 );
	    Vprod(vv[i], vv[k], vp);
	    ww[i] = Vskal(vp, vp);

	    if (ww[i] > ww[kk])
	       kk = i;
	 }

	 k = kk;
      }
      else
	 k = 0;

      fprintf( fpp, "// ppatch %d\n", np);  

      if (np=3)   /* nur Dreiecke */
      {
                     
        

      fprintf( fpp, "smooth_triangle{\n");
      for (i = 0;  i < np;  i++)
      {
         fprintf(fpp, "<");
	 for (j = 0;  j < DIM;  j++)
	 {
	    if (j == 0) { jj = 0; }
	    if (j == 1) { jj = 2; }
	    if (j == 2) { jj = 1; }
	     
	    Form_konv( fac[k][jj], txt);
	    fprintf( fpp," %s%s", (j ? " " : ""), txt);
	    if (j < DIM-1) { fprintf( fpp, ","); } 
	 }

	 fprintf(fpp, ">,<");
	 
	 for (j = 0;  j < DIM;  j++)
	 {

	    if (j == 0) { jj = 0; }
	    if (j == 1) { jj = 2; }
	    if (j == 2) { jj = 1; }

	    Form_konv( nv[k][jj], txt);
	    fprintf( fpp, " %s", txt);
	    if (j < DIM-1) { fprintf( fpp, ","); } 
	 }

         fprintf(fpp,">");
         if (i < 2) 
         { 
            fprintf(fpp, ","); 
         }
         fprintf(fpp, "\n");
	 k = (k+1 < np ? k+1 : 0);
      }
      fprintf( fpp, "}\n");
      
      }
      
      (*nfacp)++;
   }
   else
   {
      FG_povray_triangle(3, fac, nv, nfacp);

      for (j = 0;  j < DIM;  j++)
      {
	 fac[1][j] = fac[0][j];
	 nv[1][j] = nv[0][j];
      }

      FG_povray_triangle(np - 1, &fac[1], &nv[1], nfacp);
   }
}


PRIVATE short int				 /* Anzahl generierter Facetten */
FG_gfac (nfacp)					 /* Generiere Flaechen-Facetten */
unsigned int *nfacp;
{
   short int i, k, 
	     j0 = 0, j3 = 3, eastat,
	     kp[MAXNPF], ks[MAXNPF],
	     np, rpol[MAXNPF],
	     nfac;
   OTgk kd;
   OTknoten *p;
   TMelem *e0;
   OTpol   pol[MAXNPF];
   float   fac[MAXNPF][DIM],
	   nv[MAXNPF][DIM];
#ifdef MOD_TETRA
   char   txt2[15];
   String name;

#ifdef vms
   name.len = 15;
   name.txt = txt2;
#else
#define name txt2
#endif
#endif

   Nzkomp = nfac = *nfacp = 0;
   OTS_Vqueue(1);

   for (i = 0; i < MAXNPF; i++)
   {
      kp[i] = 1;  ks[i] = 2;
   }

   while (p = OTA_szell(ktab))		 /* Suche nicht behandelte Start-Zelle */
   {
      OTS_push(p);
      e0 = TMA_ele(ktab, p->Orig, &k);
      e0->m = 0;		        /* Kubus p ist bearbeitet */
      Nzkomp++;

      if ((a_mod == 2 || a_mod == 3) && fpp)
	 fprintf( fpp, "surface Z_Komp_%d\n", Nzkomp);

      while (OTA_NSele() > 0)		/* Queue nicht leer ? */
      {
	 p  = OTS_pop();

	 if (OTA_mark(p) == T)
	 {
	    i = 0;

            while (np = OTS_Stetra(p, &i, pol))
            {
	       if (np < 3)
		  printf("!! Fehler (FG_gfac): T-Facette entartet ?! np=%d\n", np);
	       else if (np > MAXNPF )
		  printf("!! Fehler (FG_gfac): T-Facette zu gross! np=%d\n", np);
	       else
	       {
		  if ((np = FG_fac(np, pol, rpol, fac, nv)) > 2)
		  {
#ifdef KONDAR
		     if (a_mod == 1)
			if ((k = KMSdfc(&j0, &j3, &np, &j0, &j0, kp, ks, fac, &j0)) == nfac+1)
			   nfac++;
#endif

		     if (a_mod == 2)
			FG_ppatch(np, fac, nv, nfacp);
		     if (a_mod == 3)
			FG_patch(np, fac, nv, nfacp);
		     if (a_mod == 4)
			FG_povray_triangle(np, fac, nv, nfacp);
		  }
		  else
		     printf("!! Warnung (FG_gfac): Verdichtete T-Facette entartet ?! np=%d\n", np);
	       }
            }
#ifdef MOD_TETRA
#ifdef KONDAR
	    KMSext();
	    sprintf( txt2, "t%02d%02d%02d#", p->Orig[0], p->Orig[1], p->Orig[2] );

	    if (k = KMSaus(&name, &eastat))
	       printf("!! Fehler bei Modellausgabe: %d\n", eastat );

	    KMSini();
#endif
#else
	 }
	 else
	 {
	    kd = OTA_gkd(p);
	    np = OTS_Spol(p, 1, kd, pol);

	    if (np < 3)
	       printf("!! Fehler (FG_gfac): W-Facette entartet ?! np=%d\n", np);
	    else if (np > MAXNPF )
	       printf("!! Fehler (FG_gfac): W-Facette zu gross! np=%d\n", np);
	    else
	    {
	       if ((np = FG_fac(np, pol, rpol, fac, nv)) > 2)
	       {
#ifdef KONDAR
		  if (a_mod == 1)
		     if ((k = KMSdfc(&j0, &j3, &np, &j0, &j0, kp, ks, fac, &j0)) == nfac+1)
			nfac++;
#endif

		   if (a_mod == 2)
			FG_ppatch(np, fac, nv, nfacp);
		   if (a_mod == 3)
			FG_patch(np, fac, nv, nfacp);
		   if (a_mod == 4)
			FG_povray_triangle(np, fac, nv, nfacp);
	       }
	       else
		  printf("!! Warnung (FG_gfac): Verdichtete W-Facette entartet ?! np=%d\n", np);
	    }
#endif 
	 }
      }						 /* Ende while */

      j3++;
   }						 /* Ende while */

   OTS_Vqueue(0);
   return ( nfac );
}

#ifdef NCHECK   /* nur zum Debuggen ... */
FG_Ncheck()
{
   char txt[50], *tp;
   OTgr r;
   OTgk kd, kdn, Orig[3];
   OTknoten *p, *pn;

   printf("-- Nachbarschaftstests (j/n) ?: ");

   if ((tp = fgets(txt,10,stdin)) && strlen(tp) > 1 && *tp == 'j')
   {
      repeat
      {
	 printf("-- [Orig,] Richtung : ");

	 if (!((tp = fgets(txt,50,stdin)) && strlen(tp)-1 > 0))
	    break;

	 if (strchr(tp, ','))
	 {
	    Orig[0] = atoi(tp);    tp = strchr(tp, ',') + 1;
	    Orig[1] = atoi(tp);    tp = strchr(tp, ',') + 1;
	    Orig[2] = atoi(tp);    tp = strchr(tp, ',') + 1;
	 }

	 r = atoi(tp);
	 p = OTA_elem(Orig);
	 kd = OTA_gkd(p);
	 printf("Knoten: (%d %d %d; %d)\n", p->Orig[0], p->Orig[1], p->Orig[2], kd);

	 if (OTA_or(r))
	    pn = OTA_gsn(p,r);
	 else if (OTA_odr(r))
	    pn = OTA_gen(p,r);
	 else if (OTA_dr(r))
	    pn = OTA_gcn(p,r);

	 if (pn)
	 {
	    kdn = OTA_gkd(pn);
	    printf("%d-Nachbar: (%d %d %d; %d)\n", r, pn->Orig[0], pn->Orig[1], pn->Orig[2], kdn);
	 }
	 else
	    printf("Nachbar nicht vorhanden!\n");
      }
   }
}
#endif
