#include <math.h>

#ifndef TRUE
#define TRUE	(1)
#define FALSE	(0)
#endif

#ifndef EPSD
#define EPSD 1E-10
#endif

#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifdef vms
#define Cabs Cabs_
#endif

void   Vprod(), Vlppg(), Vmmul();
double Vabs(), Vskal();
short  Vident();

double qakt[4],	     /* aktuell gespeicherte Quaternion  */
       Q[3][3];	     /* aktuelle Drehmatrix zu qakt      */

double Cabs(z)
double z[];
{
   return ( hypot(z[0],z[1]) );
}

double Carg(z)
double z[];
{
   return ( atan2(z[1],z[0]) );
}

void Cmul(z1,z2,zp)
double z1[], z2[], zp[];
{
   double a, b;

   a = z1[0] * z2[0] - z1[1] * z2[1];
   b = z1[0] * z2[1] + z1[1] * z2[0];
   zp[0] = a;  zp[1] = b;
}

int Cdiv(z1,z2,zq)
double z1[], z2[], zq[];
{
   double a, b;

   b = z2[0] * z2[0] + z2[1] * z2[1];

   if ( b > EPSD )
   {
      a = ( z1[0] * z2[0] + z1[1] * z2[1] ) / b;
      b = ( z1[1] * z2[0] - z1[0] * z2[1] ) / b;
      zq[0] = a;  zq[1] = b;
      return ( TRUE );
   }
   else
      return ( FALSE );
}

void Qqnm(q,A)			    /* Konversion: Quaternion -> Matrix   */
double q[], A[][3];
{
   short i;
   double q2[4], q01, q02, q03, q12, q13, q23;

   for ( i = 0; i < 4; i++ )
      q2[i] = q[i] * q[i];

   q01 = 2.0 * q[0] * q[1];
   q02 = 2.0 * q[0] * q[2];
   q03 = 2.0 * q[0] * q[3];
   q12 = 2.0 * q[1] * q[2];
   q13 = 2.0 * q[1] * q[3];
   q23 = 2.0 * q[2] * q[3];

   A[0][0] = 1.0 - 2 * (q2[2] + q2[3]);
   A[1][1] = 1.0 - 2 * (q2[1] + q2[3]);
   A[2][2] = 1.0 - 2 * (q2[1] + q2[2]);
   A[0][1] = q12 - q03;
   A[0][2] = q13 + q02;
   A[1][0] = q12 + q03;
   A[1][2] = q23 - q01;
   A[2][0] = q13 - q02;
   A[2][1] = q23 + q01;
}

void Qmnq(A,q)			    /* Konversion: Matrix -> Quaternion   */
double q[], A[][3];
{
   double tr, s;

#define nxt(i) ((i+1)%3)

   if (( tr = A[0][0] + A[1][1] + A[2][2] ) > 0.0 )
   {
      s = sqrt( tr + 1.0 );
      q[0] = s * 0.5;
      s    = 0.5 / s;
      q[1] = ( A[2][1] - A[1][2]) * s;
      q[2] = ( A[0][2] - A[2][0]) * s;
      q[3] = ( A[1][0] - A[0][1]) * s;
   }
   else
   {
      short  i, j, k;

      i = ( A[1][1] > A[0][0] ? 1 : 0 );

      if ( A[2][2] > A[i][i] )
	 i = 2;

      j = nxt(i);  k = nxt(j);
      s = sqrt((A[i][i] - (A[j][j]+A[k][k])) + 1.0 );
      q[i+1] = s * 0.5;
      s      = 0.5 / s;
      q[0]   = ( A[k][j] - A[j][k]) * s;
      q[j+1] = ( A[j][i] + A[i][j]) * s;
      q[k+1] = ( A[k][i] + A[i][k]) * s;
   }
}

void Qmul(q1,q2,qp)		    /* Multiplikation zweier Quaternionen */
double q1[], q2[], qp[];
{
   double v[3];
   short i;

   qp[0] = q1[0] * q2[0] - Vskal(&q1[1],&q2[1]);
   Vprod(&q1[1],&q2[1],v);

   for ( i = 1; i < 4; i++ )
      qp[i] = q1[0] * q2[i] + q2[0] * q1[i] + v[i-1];
}

void Qmulv(v,q,vq)		    /* Multiplikation von Quaternion und Vektor */
double v[], q[], vq[];
{
   if ( Vident(4,q,qakt))
   {
      short i;

      for ( i = 0; i < 4; i++ )
	 qakt[i] = q[i];

      Qqnm(q,Q);
   }

   Vmmul(Q,v,vq);	   /* vq := Q*v  */
}

void Qdiv(q1,q2,qr)		    /* Division zweier Einheits-Quaternionen */
double q1[], q2[], qr[];
{
   double qq[4];
   short i;

   qq[0] = q2[0];

   for ( i = 1; i < 4; i++ )
      qq[i] = -q2[i];

   Qmul(q1,qq,qr);
}

void Qdivv(v,q,vq)		    /* Division Vektor / Einheits-Quaternion */
double v[], q[], vq[];
{
   double qq[4];
   short i;

   qq[0] = q[0];

   for ( i = 1; i < 4; i++ )
      qq[i] = -q[i];

   Qmulv(v,qq,vq);
}

void Qpow(q,e,qp)
double q[], qp[], e;
{
   double cw, sw, w;
   short i;

   cw = q[0];  sw = Vabs(&q[1]);

   if ( abs(sw) > EPSD )
      for ( i = 1; i < 4; i++ )
         qp[i] = q[i] / sw;
 
   w = e * acos(cw);  sw = sin(w);
   qp[0] = cos(w);

   for ( i = 1; i < 4; i++ )
      qp[i] *= sw;
}

short Vident(n,v1,v2)
short  n;
double v1[], v2[];
{
   short i;

   for ( i = 0; i < n; i++ )
      if ( fabs(v1[i]-v2[i]) > EPSD )
	 return ( 1 );

   return ( 0 );
}

void Vadd(v1,v2,vs)
double v1[], v2[], vs[];
{
   short i;

   for ( i = 0; i < 3; i++ )
      vs[i] = v1[i] + v2[i];
}

void Vdif(v1,v2,vd)
double v1[], v2[], vd[];
{
   short i;

   for ( i = 0; i < 3; i++ )
      vd[i] = v1[i] - v2[i];
}

double Vskal(v1,v2)
double v1[], v2[];
{
   double h;
   short  i;

   for ( h = .0, i = 0; i < 3; i++ )
      h += v1[i] * v2[i];

   return ( h );
}

double Vabs(v)
double v[];
{
   return ( sqrt(Vskal(v,v)) );
}

void Vricht(v1,v2)
double v1[], v2[];
{
   short  i;
   double b;

   if (( b = Vabs(v1)) > .0 )
      for ( i = 0; i < 3; i++ )
         v2[i] = v1[i] / b;
}

void Vprod(v1,v2,vp)
double v1[], v2[], vp[];
{
   vp[0] = v1[1] * v2[2] - v1[2] * v2[1];
   vp[1] = v1[2] * v2[0] - v1[0] * v2[2];
   vp[2] = v1[0] * v2[1] - v1[1] * v2[0];
}

double Vspat(v1,v2,v3)
double v1[], v2[], v3[];
{
   double vh[3];

   Vprod(v1,v2,vh);
   return ( Vskal(v3,vh));
}

double Vnor( kmod, v1, v2, v3, vn )
short kmod;
double v1[], v2[], v3[], vn[];
{
   short i;
   double d, vd1[3], vd3[3];

   if ( kmod <= 1 )
   {
      Vdif( v1, v2, vd1 );	 /* vd1 := v1 - v2   */
      Vdif( v3, v2, vd3 );	 /* vd3 := v3 - v2   */
      Vprod( vd1, vd3, vn );	 /* vn  := vd1 x vd3 */
   }
   else
      Vprod( v1, v3, vn );	 /* vn  := v1 x v3 */

   d = Vabs(vn);

   if ( kmod % 2 == 1 && d > .0 )
   {
      for ( i = 0; i < 3; i++ )
	 vn[i] = vn[i] / d;

      d = 1.0;
   }

   return ( d );
}

void Vmmul(A,x,y)	      /* y := A*x  */
double A[][3], x[], y[];
{
   short i, j;

   for ( j = 0; j < 3; j++ )
   {
      y[j] = 0.0;

      for ( i = 0; i < 3; i++ )
	 y[j] += A[j][i] * x[i];
   }
}

double Vdist(v1,v2)
double v1[], v2[];
{
   double vh[3];

   Vdif(v1,v2,vh);
   return ( Vabs(vh));
}

double Vdpg(v1,vg,vr)
double v1[], vg[], vr[];
{
   double vh[3];

   Vlppg(v1,vg,vr,vh);
   return ( Vdist(v1,vh));
}

double Vdpe(v1,e)
double v1[], e[];
{
   return ( Vskal(v1,e) - e[3] );
}

double Vdge(vg,vr,e)
double vg[], vr[], e[];
{
   if ( fabs(Vskal(vr,e)) > EPSD )
      return ( .0 );
   else
      return ( Vskal(vg,e) - e[3] );
}

double Vdee(e1,e2)
double e1[], e2[];
{
   if ( fabs(fabs(Vskal(e1,e2)) - 1.0 ) > EPSD )
      return ( .0 );
   else
      return ( e2[3] - e1[3] * Vskal(e1,e2));
}

double Vdgg(vg1,vr1,vg2,vr2)
double vg1[], vr1[], vg2[], vr2[];
{
   double h1[3], h2[3], h;
   short i;

   Vprod(vr1,vr2,h1);	      /* h1 := vr1 x vr2    */
   Vdif(vg2,vg1,h2);	      /* h2 := vg2 - vg1    */

   if (( h = Vabs(h1)) > EPSD )
   {			      /* die Geraden sind windschief */
      for ( i = 0; i < 3; i++ )
	 h1[i] /= h;

      return ( fabs(Vskal(h1,h2)));
   }
   else			      /* die Geraden sind parallel   */
      return ( Vdpg(vg1,vg2,vr2));
}

double Vwin(v1,v2)
double v1[], v2[];
{
   double w, vr1[3], vr2[3];

   Vricht(v1,vr1);  Vricht(v2,vr2);
   w = Vskal(vr1,vr2);
   w = ( w > 1.0 ? 1.0 : ( w < -1.0 ? -1.0 : w ));
   return ( acos(w));
}

double Vwpe(v,e)
double v[], e[];
{
   double w, vr[3];

   Vricht(v,vr);
   w = fabs(Vskal(vr,e));
   w = ( w > 1.0 ? 1.0 : w );
   return ( asin(w));
}

short Vspge(vg,vr,en,ed,dp)     /* Durchstosspunkt existiert: 1/0 */
double vg[], vr[], en[], ed, dp[];
{
   double fn, f;
   short  i;

   if ( fabs(fn = Vskal(vr,en)) < EPSD )
      return ( 0 );
   else
   {
      f = ( ed - Vskal(vg,en)) / fn;

      for ( i = 0; i < 3; i++ )
	 dp[i] = vg[i] + f * vr[i];
   }

   return ( 1 );
}

short Vspgg(vg1,vr1,vg2,vr2,sp)	    /* Schnittpunkt existiert: 1/0 */
double vg1[], vr1[], vg2[], vr2[], sp[];
{
   double n1[3], n2[3], d;

   Vprod(vr1,vr2,n1);	      /* n1 := vr1 x vr2    */
   Vprod(vr1,n1,n2);	      /* n2 := vr1 x n1     */
   d = Vskal(vg1,n2);	      /* n2 := vg1 * n2     */

   return ( Vspge(vg2,vr2,n2,d,sp));
}

void Vve(v1,v2,v3,e)
double v1[], v2[], v3[], e[];
{
   double h;

   h = Vnor(1,v1,v3,v2,e);
   e[3] = Vskal(v1,e);
}

void Vvepg(v1,vg,vr,e)
double v1[], vg[], vr[], e[];
{
   double vh[3];

   Vadd(vg,vr,vh);
   Vve(v1,vg,vh,e);
}

void Vvegg(vg1,vr1,vg2,vr2,e)
double vg1[], vr1[], vg2[], vr2[], e[];
{
   double v1[3], v2[3], v3[3];

   Vspgg(vg1,vr1,vg2,vr2,v1);
   Vadd(v1,vr1,v2);
   Vadd(v1,vr2,v3);
   Vve(v1,v2,v3,e);
}

void Vlppg(v1,vg,vr,vl)	      /* vl := Lotfusspunkt von v1 auf (vg,vr)  */
double v1[], vg[], vr[], vl[];
{
   short  i;
   double h, vh[3];

   Vdif(v1,vg,vl);
   h = Vskal(vr,vl);

   for ( i = 0; i < 3; i++ )
      vh[i] = h * vr[i];

   Vadd(vg,vh,vl);
}

void Vlppe(v1,e,vl)	      /* vl := Lotfusspunkt von v1 auf Ebene e  */
double v1[], e[], vl[];
{
   short  i;
   double h;

   h = Vdpe(v1,e);

   for ( i = 0; i < 3; i++ )
      vl[i] = v1[i] - h * e[i];
}

void Vlpgg(vg1,vr1,vg2,vr2,vl)	      /* vl := Lotfusspunkt von (vg1,vr1) auf (vg2,vr2)  */
double vg1[], vr1[], vg2[], vr2[], vl[];
{
   double vh[3], e[4];

   Vprod(vr1,vr2,vh);
   Vadd(vg1,vh,vh);
   Vvepg(vh,vg1,vr1,e);
   Vspge(vg2,vr2,e,e[3],vl);
}

void Vlgge(vg,vr,e,vgl,vrl)	      /* vgl,vrl := Lotgerade von (vg,vr) auf Ebene e   */
double vg[], vr[], e[], vgl[], vrl[];
{
   short  i;
   double s, v1[3], v2[3];

   Vadd(vg,vr,v2);
   Vlppe(vg,e,v1);
   Vlppe(v2,e,v2);
   Vdif(v2,v1,vrl);
   Vricht(vrl,vrl);
   s = Vskal(v1,vrl);

   for ( i = 0; i < 3; i++ )
      vgl[i] = v1[i] - s * vrl[i];
}

void Vlrpp(v1,v2,lr)	      /* lr := Lotrichtung von v1 auf v2  */
double v1[], v2[], lr[];
{
   Vdif(v2,v1,lr);
   Vricht(lr,lr);
}

void Vlrpe(v,e,lr)	      /* lr := Lotrichtung von v auf e   */
double v[], e[], lr[];
{
   short i;

   if ( Vdpe(v,e) > .0 )
      for ( i = 0; i < 3; i++ )
	 lr[i] = -e[i];
   else
      for ( i = 0; i < 3; i++ )
	 lr[i] = e[i];
}

void Vlrpg(v,vg,vr,lr)	      /* lr := Lotrichtung von v auf (vg,vr)  */
double v[], vg[], vr[], lr[];
{
   double vl[3];

   Vlppg(v,vg,vr,vl);
   Vdif(vl,v,lr);
   Vricht(lr,lr);
}

void Vlrgg(vg1,vr1,vg2,vr2,lr)	 /* lr := Lotrichtung von (vg1,vr1) auf (vg2,vr2)  */
double vg1[], vr1[], vg2[], vr2[], lr[];
{
   short i;
   double vh[3], vd[3];

   Vprod(vr1,vr2,vh);
   Vdif(vg2,vg1,vd);

   if ( Vabs(vh) > EPSD )
   {
      Vricht(vh,lr);

      if ( Vskal(vd,lr) < .0 )
	 for ( i = 0; i < 3; i++ )
	    lr[i] = -lr[i];
   }
   else
      Vricht(vd,lr);
}

void Vkaku(v1,u,v,r)
double v1[],*u,*v,*r;
{
   double q;

   q = v1[0] * v1[0] + v1[1] * v1[1];

   if ( q >= EPSD )
   {
      *r = sqrt(q+v1[2]*v1[2]);
      q = sqrt(q);
      *v = atan(v1[2]/q);
      *u = atan2(v1[1],v1[0]);

      if ( *u < .0 )
	 *u += 2 * PI;
   }
   else
   {
      *u = .0;
      *r = fabs(v1[2]);

      if ( *r > EPSD )
      {
	 *v = PI / 2;

	 if ( v1[2] < .0 )
	    *v = -*v;
      }
      else
      {
	 *v = .0;  *r = .0;
      }
   }
}

void Vkuka(u,v,r,vr)
double u,v,r,vr[];
{
   vr[0] = vr[1] = cos(v);
   vr[2] = sin(v) * r;
   vr[0] *= r * cos(u);
   vr[1] *= r * sin(u);
}

short Vlgs3(A,x)     /* Vlgs3 = 1: x := Loesung des LGS Ax = b (mit b = A(4,.) */
double A[][4], x[];
{
   short i, j, k;
   double p, B[3][4];

   for ( j = 0; j < 3; j++ )
      for ( i = 0; i < 4; i++ )
	 B[j][i] = A[j][i];

   for ( k = 0; k < 3; k++ )
   {
      j = k;  p = fabs(B[k][k]);

      for ( i = k+1; i < 3; i++ )
	 if ( p < fabs(B[i][k]))
	    p = fabs(B[j=i][k]);

      if ( p < EPSD )
	 return ( 0 );		 /* LGS ist singulaer */

      if ( j != k )
	 for ( i = 0; i < 4; i++ )
	 {
	    p = B[j][i];  B[j][i] = B[k][i];  B[k][i] = p;
	 }

      p = 1.0 / B[k][k];

      for ( j = k+1; j < 4; j++ )
	 B[k][j] *= p;

      for ( i = k+1; i < 3; i++ )
      {
	 for ( j = k+1; j < 4; j++ )
	    B[i][j] -= B[i][k] * B[k][j];

	 B[i][k] *= p;
      }
   }

   for ( k = 2; k >= 0; k-- )
   {
      p = B[k][3];

      for ( j = k+1; j < 3; j++ )
	 p -= B[k][j] * x[j];

      x[k] = p;
   }

   return ( 1 );
}
