/*
**
**  INCLUDE FILES
**
*/

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include "fxdef.h"
#include "tmdef.h"
#include "ot.h"


/*
**
**  MACRO DEFINITIONS
**
*/

#define ECKEN  0
#define INNEN  1
#define SEITEN 2

#define MAXKNSET    100	    /* maximale Knotenmenge fuer OTS_Spol  */
#define MINKD	    2	    /* minimale Koordinatendistanz fuer Wuerfel */

#define SIGNUM(w) (w > 0.0 ? 1 : -1)
#define	SigEQ(w1,w2)	(w1 > .0 && w2 > .0 || w1 < .0 && w2 < .0)

#define OTI_asspk(p,k)	{short I;for (I=0;I<DIM;I++) p[I]=otw.Orig[I]+epsd2*k[I];}
    /*  OTI_asspk: absolute Punktkoordinaten aus Gitterkoordinaten berechnen */
#define OTI_assgk(p1,p2)	{short _i;for(_i=0;_i<DIM;_i++) p1[_i]=p2[_i];}
#define OTI_ccomp(c)		(OTA_dr(c) ? OTccomp[c] : XXX)

OTwurzel otw;
OTgr OTccomp[] = { UVW, XVW, UVZ, XVZ, UYW, XYW, UYZ, XYZ }, 
     OTscomp[] = { UU, XX, VV, YY, WW, ZZ },
     OTrref[6][GRD] = {{ UYZ, XYZ, UYW, XYW, UVZ, XVZ, UVW, XVW },
		       { XVZ, UVZ, XVW, UVW, XYZ, UYZ, XYW, UYW },
		       { XYW, UYW, XYZ, UYZ, XVW, UVW, XVZ, UVZ },
		       { UVZ, XVZ, UVW, XVW, UYZ, XYZ, UYW, XYW },
		       { UYW, XYW, UYZ, XYZ, UVW, XVW, UVZ, XVZ },
		       { XVW, UVW, XVZ, UVZ, XYW, UYW, XYZ, UYZ }},
     OTderg[GRD][6] = {{ XX, YY, ZZ, XY, YZ, XZ },	       /* Ergaenzungsrichtungen zu einer Diagonalrichtung */
		       { UU, YY, ZZ, UY, YZ, UZ },
		       { XX, YY, WW, XY, YW, XW },
		       { UU, YY, WW, UY, YW, UW },
		       { XX, VV, ZZ, XV, VZ, XZ },
		       { UU, VV, ZZ, UV, VZ, UZ },
		       { XX, VV, WW, XV, VW, XW },
		       { UU, VV, WW, UV, VW, WW }},
     OToderg[12][2] = {{ XX, YY }, { UU, YY }, { UU, VV },     /* Ergaenzungsrichtungen zu einer Ortho-Diagonalrichtung */
		       { XX, VV }, { XX, ZZ }, { XX, WW },
		       { UU, WW }, { UU, ZZ }, { YY, ZZ },
		       { VV, ZZ }, { VV, WW }, { YY, WW }},
     OTdror[6][4]   = {{ XYZ, XYW, XVZ, XVW },		       /* Diagonalrichtungen zu einer Orthogonal-Richtung */
		       { UYZ, UYW, UVZ, UVW },
		       { XYZ, UYZ, XYW, UYW },
		       { XVZ, UVZ, XVW, UVW },
		       { XYZ, UYZ, XVZ, UVZ },
		       { XYW, UYW, XVW, UVW }},
     OTdrodr[12][2] = {{ XYZ, XYW }, { UYZ, UYW },	       /* Diagonalrichtungen zu einer Orthodiagonalgonal-Richtung */
		       { UVZ, UVW }, { XVZ, XVW },
		       { XYZ, XVZ }, { XYW, XVW },
		       { UYW, UVW }, { UYZ, UVZ },
		       { XYZ, UYZ }, { XVZ, UVZ },
		       { XVW, UVW }, { XYW, UYW }},
     OTrstr[] = { XW, YW, UW, VW, XV, XZ,
		  YZ, UZ, VZ, XY, UY, UV };		       /* Richtung der Kubus-Strecken */
     OTrfac[] = { XX, YY, UU, VV, WW, ZZ };		       /* Richtung der Kubus-Facetten */

PRIVATE short ir = 0,					       /* aktueller Riss */
   kt[3][3] = {{ 0, 1, 2 }, { 1, 2, 0 }, { 2, 0, 1 }},	       /* Achsen-Anordnung bei GR, AR, KR */
   OTirel3[7]     = { 0, 9, 14, 18, 20, 23, 24 },	       /* Punktrelationen fuer eindeutigen Schnitt bei 35/53 */
   OTprel3[24][2] = {{ 1, 2 }, { 1, 3 }, { 1, 4 }, { 1, 5 }, { 2, 3 },
		     { 3, 4 }, { 3, 7 }, { 4, 5 }, { 4, 7 }, { 2, 3 },
		     { 2, 5 }, { 2, 6 }, { 4, 5 }, { 5, 6 }, { 3, 6 },
		     { 3, 7 }, { 5, 6 }, { 6, 7 }, { 4, 7 }, { 6, 7 },
		     { 5, 6 }, { 6, 7 }, { 5, 7 }, { 6, 7 }},
   OTirel4[6]     = { 0, 16, 21, 24, 25, 26 },		       /* Punktrelationen fuer eindeutigen Schnitt bei 44 */
   OTprel4[26][3] = {{ 1, 2, 3 }, { 1, 2, 4 }, { 1, 2, 5 }, { 1, 2, 6 }, { 1, 3, 4 },
		     { 1, 3, 7 }, { 1, 4, 5 }, { 1, 4, 7 }, { 1, 5, 6 }, { 2, 3, 7 },
		     { 3, 4, 7 }, { 3, 2, 6 }, { 3, 6, 7 }, { 4, 5, 6 }, { 4, 5, 7 },
		     { 4, 6, 7 }, { 2, 3, 5 }, { 2, 5, 6 }, { 2, 6, 7 }, { 4, 5, 6 },
		     { 5, 6, 7 }, { 3, 6, 7 }, { 4, 6, 7 }, { 5, 6, 7 }, { 4, 6, 7 },
		     { 5, 6, 7 }},
   OTcstr[] = { 7, 8, 5, 6, 10, 2, 3, 0, 1, 11, 4, 9 },	       /* Index der Gegenkante zu einer Wuerfelstrecke */
   OTcfac[] = { 2, 3, 0, 1, 5, 4 },			       /* Index der Gegenfacette zu einer Wuerfelfacette */
   OTgkkub[] = { 2, 1, 0, 3 },				       /* Index der Gegenkante zu einer Wuerfelfacette */
   OTgkfac[] = { 2, 3, 0, 1 };				       /* Index der Gegenkante zu einer Nachbarfacette */

typedef struct
{
   OTknoten *kn;
   short    done;
} OTknset;

typedef
struct strecke			 /* Metrik und Topologie einer Strecke */
{
   short    kv;			 /* Kennung der variierenden Achse */
   OTgk	    pa[DIM], pe[DIM];	 /* Endpunkte in Gitterkoordinaten */
   OTpunkt  wa, we;		 /* Funktionswerte in den Endpunkten */
} Strecke;

typedef
struct facette			 /* Metrik und Topologie einer Facette */
{
   short    str[4];		 /* Strecken: Kennung der variierenden Achse */
   OTgk	    pkt[4][DIM];	 /* Eckpunkte in Gitterkoordinaten */
   OTpunkt  w[4];		 /* Funktionswerte in den Eckpunkten */
} Facette;

typedef
struct tetraeder		 /* Metrik und Topologie des Tetraeders: */
{
   short    str[6][2],		 /* Strecken: Index von Anfangs-, Endpunkt */
	    psr[4][3],		 /* Punkt-Strecken-Relation: Indizes der Strecken, die an einen Punkt anschliessen */
	    fac[4][3],		 /* Facetten: Nummern (!) der durchlaufenen orientierten Strecken */
	    sfr[6][2],		 /* Strecken-Facetten-Relation: Indizes der Facetten (li,re), die an eine Strecke anschliessen */
	    skr[6][2];		 /* Strecken-Kanten-Relation: Kanten-Index einer Strecke bzgl. ihrer (li,re) Nachbarfacette*/
   OTgk	    pkt[4][DIM];	 /* Eckpunkte in Gitterkoordinaten */
   OTpunkt  w[4];		 /* Funktionswerte in den Eckpunkten */
   OTknoten *p;			 /* Adresse des zugehoerigen OT-Knotens */
} Tetraeder;

typedef
struct kubus			 /* Metrik und Topologie des Wuerfels: */
{
   short    str[12][3],		 /* Strecken: Index von Anfangs-, Endpunkt, variierender Achse */
	    psr[8][3],		 /* Punkt-Strecken-Relation: Indizes der Strecken, die an einen Punkt anschliessen */
	    fac[6][4],		 /* Facetten: Nummern (!) der durchlaufenen orientierten Strecken */
	    sfr[12][2],		 /* Strecken-Facetten-Relation: Indizes der Facetten (li,re), die an eine Strecke anschliessen */
	    skr[12][2];		 /* Strecken-Kanten-Relation: Kanten-Index einer Strecke bzgl. ihrer (li,re) Nachbarfacette*/
   OTgk	    pkt[8][DIM];	 /* Eckpunkte in Gitterkoordinaten */
   OTpunkt  w[8];		 /* Funktionswerte in den Eckpunkten */
   OTknoten *p;			 /* Adresse des zugehoerigen OT-Knotens */
} Kubus;

PRIVATE Kubus kub = {
		      {{ 0, 1, 1 }, { 1, 2, 0 }, { 2, 3, 1 }, { 0, 3, 0 },	    /* Strecken: */
		       { 0, 4, 2 }, { 4, 5, 1 }, { 5, 6, 0 }, { 6, 7, 1 },
		       { 4, 7, 0 }, { 1, 5, 2 }, { 2, 6, 2 }, { 3, 7, 2 }},
		      {{ 0, 3, 4 }, { 0, 1, 9 }, { 1, 2, 10 }, { 2, 3, 11 },	    /* Punkt-Strecken-Relation */
		       { 4, 5, 8 }, { 5, 6, 9 }, { 5, 6, 9 }, { 6, 7, 10 }},
		      {{ 5, 6, -10, -1 }, { 10, 7, -11, -2 }, { 11, 8, -12, -3 },   /* Facetten: */
		       { 12, -9, -5, 4 }, { 1, 2, 3, -4 }, { -8, -7, -6, 9 }},
		      {{ 0, 4 }, { 1, 4 }, { 2, 4 }, { 4, 3 },			    /* Strecken-Facetten-Relation */
		       { 3, 0 }, { 5, 0 }, { 5, 1 }, { 5, 2 },
		       { 3, 5 }, { 0, 1 }, { 1, 2 }, { 2, 3 }},
		      {{ 3, 0 }, { 3, 1 }, { 3, 2 }, { 3, 3 },			    /* Strecken-Kanten-Relation */
		       { 2, 0 }, { 2, 1 }, { 1, 1 }, { 0, 1 },
		       { 1, 3 }, { 2, 0 }, { 2, 0 }, { 2, 0 }} 
		     };

PRIVATE Tetraeder tet = {
			 {{ 0, 1 }, { 1, 2 }, { 0, 2 }, { 0, 3 }, { 1, 3 }, { 2, 3 }},	  /* Strecken */
			 {{ 0, 2, 3 }, { 0, 1, 4 }, { 1, 2, 5 }, { 3, 4, 5 }},		  /* Punkt-Strecken-Relation */
			 {{ 4, -5, -1 }, { 5, -6, -2 }, { 6, -4, 3 }, { 1, 2, -3 }},	  /* Facetten */
			 {{ 0, 3 }, { 1, 3 }, { 3, 2 }, { 2, 0 }, { 0, 1 }, { 1, 2 }},	  /* Strecken-Facetten-Relation */
			 {{ 2, 0 }, { 2, 1 }, { 2, 2 }, { 1, 0 }, { 1, 0 }, { 1, 0 }}	  /* Strecken-Kanten-Relation */
			};

PRIVATE Facette fac;
PRIVATE Strecke str;

PRIVATE FLEX *FXkub;

PRIVATE unsigned int Nqueue, OTnknot = 0;

short int  OTI_ksect ();		/* forward */
OTknoten * OTA_gen ();
OTknoten * OTA_gsn ();
void	  OTI_vw3 ();
OTgr	  OTI_class ();
void	  OTI_rcomp ();
OTknoten * OTI_gen ();
OTknoten * OTI_gsn ();
OTknoten * OTI_gcn ();
short int  OTI_upd ();
short int  OTS_Spol ();
void	  OTI_dkub ();
void	  OTI_dfac ();
void	  OTI_dstr ();
short int  OTI_kident ();
short int  OTI_nxtfk ();
short int  OTI_nTfk ();
void	  OTI_conTk ();
void	  OTI_kTsect ();
void	  OTI_Sp ();
void	  OTI_Spnorth ();
void	  OTI_rckn ();
void	  OTI_gkoord ();
OTknoten * OTI_skn ();
short	  OTI_planp ();
short	  OTI_pland ();
short	  OTI_planc ();
void	  OTI_Ngrad ();
void	  OTI_Nvekt ();
OTpunkt	  OTI_NAdist ();
void	  OTI_Exit ();

#ifdef XXX	/* z.Zt. nicht benoetigt */
OTknoten * OTA_Mnachbar ();
OTpunkt  * OTI_Melem  ();
#endif

extern TMtab *ktab, *ptab;
extern OTpunkt epsd2, epsd3, epsdp, epsdd, epsdc;
extern double F_val();
extern short pchk, dchk, cchk, schk, tcheck,
	 nv_mod,		       /* Normalvektoren der Schnittpunkte speichern */
	 Iter, Iter1, Iter2;

void
OTS_ini (Orig, sl, epsd, dmax, lmax, rmax)   /* Initialisiere einen Octree */
OTpunkt Orig[],				 /* (E) Linke untere Ecke */
	sl,				 /* (E) Seitenlaenge */
	epsd;				 /* (E) max. Aufloesung */
short int dmax,				 /* (E) max. OT-Tiefe fuer Detail-Erkennung */
	  lmax,				 /* (E) max. Level */
	  rmax;				 /* (E) max. Level fuer regulaere Gebiete */
{
   otw.start = NULL;
   otw.sl    = sl;
   otw.epsd  = epsd;
   otw.mlv1  = dmax;
   otw.mlev  = lmax;
   otw.mlvr  = rmax;
   otw.kd1max = ( 1 << lmax - dmax );
   OTI_assgk( otw.Orig, Orig );
}

OTknoten *
OTS_neu (void)
{
   short int i;
   OTknoten *p;

   if (p = malloc(sizeof(OTknoten)))
   {
      for (i = 0;  i < GRD;  i++)
	 p->n.nxt[i] = NULL;

      p->vorg = NULL;
      OTnknot++;
   }

   return (p);
}

void
OTS_gkv(e)		      /* Belegen des TM-Wertes fuer ein neues Element */
TMelem *e;		      /* (E) Verweis auf TM-Struktur */
{
   OTpunkt  p[DIM];

   OTI_asspk(p,e->k);
   e->w = F_val(p);
}

void
OTS_ins (p, q, r)	      /* Fuege Knoten ein */
OTknoten *p,		      /* (E) Knoten, unter dem einzufuegen ist */
	 *q;		      /* (E) einzufuegender Knoten */
OTgr r;			      /* (E) Richtung, in der einzufuegen ist */
{
   short int k;
   OTgk kd;

   k = otw.mlev - ( q->level = p->level + 1 );
   kd = ( 1 << k );
   OTI_assgk( q->Orig, p->Orig );

   switch (r)
   {
      case XYZ:
	 q->Orig[0] += kd;
      case UYZ:
	 q->Orig[1] += kd;
      case UVZ:
	 q->Orig[2] += kd;
	 break;
      case XVZ:
	 q->Orig[2] += kd;
      case XVW:
	 q->Orig[0] += kd;
	 break;
      case XYW:
	 q->Orig[0] += kd;
      case UYW:
	 q->Orig[1] += kd;
   }

   q->vorg = p;
   p->n.nxt[r] = q;
   p->typ = G;
}

void
OTS_del (p)		      /* Loesche einen (Teil-) Quadtree */
OTknoten *p;		      /* (E) Knoten, ab dem geloescht werden soll */
{
   if (p)
   {
      short int i;

      if ( p->typ == G )
	 for (i = 0;  i < GRD;  i++)
	    if (p->n.nxt[i])
	       OTS_del(p->n.nxt[i]);

      free(p);
      OTnknot--;
   }
}

void
OTS_ekn (p, typ)	      /* definiere Endknoten (unbehandelt) */
OTknoten *p;		      /* (E) Knoten, der markiert werden soll */
OTtyp  typ;		      /* (E) Typ des Endknotens */
{
   short int j;
   TMelem *e;

   if ( !(e = TMA_ele(ktab, p->Orig, &j)) || j != 1)
   {
      printf("!! Fehler (OTS_ekn): Element nicht gefunden!\n");
      exit();
   }

   e->m = -1;		       /* Zelle ist unbehandelt ! */
   OTS_mark(p, typ);	       /* Endknoten */
}

void
OTS_div (p, kd)		      /* Unterteile einen Knoten */
OTknoten *p;		      /* (E) Knoten, der unterteilt werden soll */
OTgk kd;		      /* (E) aktuelle Koordinaten-Differenz  */
{
   short int i;
   OTknoten *q;

   for (i = 0;  i < GRD;  i++)
   {
      if (q = OTS_neu())
	 OTS_ins(p, q, i);
      else
      {
	 int nsize;

	 nsize = OTnknot * sizeof(OTknoten);
	 printf("!! Fehler (OTS_div): Speichermangel! (# Knoten: %d, %d Byte)\n", OTnknot, nsize );
	 exit();
      }
   }

   if (p->level >= otw.mlv1)
   {			    /* Gitterpunkte der Stufe 2 aktualisieren */
      kd >>= 1;

      if (OTI_upd(ktab, p->Orig, kd) < 0)
      {
	 printf("!! Fehler (OTS_div): Speichermangel - Update nicht moeglich!\n");
	 exit();
      }
   }
}

void
OTS_part (p)		      /* Untersuche und zerlege ggf. den Knoten p */
OTknoten *p;		      /* (E) zu untersuchender Knoten */
{
   short int i, j,
	     ks, np,
	     mp, mn;
   OTgk kd;
   OTpunkt dv;
   OTknoten *q;

   kd = OTA_gkd(p);
   mn = mp = 1;
   OTI_vw3(ktab, ECKEN, p->Orig, kd, kd, &mp, &mn);

   if ( !(mp || mn))		       /* Vorzeichenwechsel */
   {
      if ( kd > otw.kd1max )	       /* bereits fein genug unterteilt ? */
      {
	 OTS_div(p, kd);
	 p->ksect = -1;
      }
      else
      {
	 OTI_dkub(p, kd);
	 ks = p->ksect = OTI_ksect(p, kd);

	 if ( ks == 0 )
	 {
	    OTS_mark(p, W);
	 }
	 else if ( ks < 0 )
	 {
	    if (kd == MINKD)	       /* Wuerfel -> Tetraeder */
	    {
	       OTI_kTsect(p);
	       OTS_ekn(p,T);
	    }
	    else
	       OTS_div(p, kd);
	 }
	 else  /* if (ks > 0)  */
	 {
	    if ( kd == MINKD )		    /* max. Tiefe erreicht */
	       OTS_ekn(p,B);
	    else
	    {
	       OTpunkt dp, nvekt[DIM];
	       OTpol	spol[6];

	       if (mp = (pchk || dchk || cchk))	     /* Signal: Endknoten */
	       {
		  dv = epsd2 * kd * 0.01;
		  np = OTS_Spol(p, 0, kd, spol );
assert( 2 < np && np < 7 );
	       }

	       if ( pchk || dchk )
	       {
		  OTI_Nvekt(np, spol, dv, nvekt);

		  if (pchk)
		     mp = OTI_planp(np, spol, nvekt);

		  if (dchk && mp)
		     mp = OTI_pland(np, spol, dv, nvekt);
	       }

	       if (cchk && mp)
		  mp = OTI_planc(np, spol, dv);

	       if (mp)
		  OTS_ekn(p,B);
	       else
		  OTS_div(p, kd);
	    }
	 }
      }
   }
   else if (!(tcheck && p->typ == G))	       /* kein Vorzeichenwechsel */
   {
      if (p->level < otw.mlv1)
      {
	 OTI_vw3(ktab, INNEN, p->Orig, kd, otw.kd1max, &mp, &mn);

	 if (mp || mn)
	    OTI_vw3(ktab, SEITEN, p->Orig, kd, otw.kd1max, &mp, &mn);

	 if (mp || mn)
	 {
	    OTS_mark(p, W);	       /* Endknoten (White) */
	 }
	 else
	 {
	    OTS_div(p, kd);   /* Unterteilen */
	    p->ksect = -1;
	 }
      }
      else
      {
	 OTS_mark(p, W);	       /* Endknoten (White) */
      }
   }

   if (p->typ == G)
      for (i = 0;  i < GRD;  i++)
	 if (q = p->n.nxt[i])
	    OTS_part(q);
}

void
OTS_join(p)		      /* Untersuche und reduziere ggf. den Knoten p */
OTknoten *p;		      /* (E) zu untersuchender Knoten */
{
   short int i, j,
	     ks, np,
	     mp, mn;
   OTgk kd;
   OTpunkt dv;
   OTknoten *q;
   TMelem *e;

   if (p->typ == G)
   {
      for (i = 0;  i < GRD;  i++)
	 if ((q = p->n.nxt[i]) && q->typ == G)
	    OTS_join(q);

      if (p->level < otw.mlvr && p->ksect < 0)
      {
         for (i = j = 0;  i < GRD && j == 0;  i++)
            if ((q = p->n.nxt[i]) && (q->typ == G || q->typ == T))
	       j++;

         if (j == 0)
	 {
	    kd = OTA_gkd(p);
	    OTI_dkub(p, kd);
	    ks = p->ksect = OTI_ksect(p, kd);

	    if (ks > 0)
	    {
	       OTpunkt dp, nvekt[DIM];
	       OTpol	spol[6];

	       if (mp = (pchk || dchk || cchk))	     /* Signal: Endknoten */
		  dv = epsd2 * kd * 0.01;

	       np = OTS_Spol(p, 0, kd, spol );
assert( 2 < np && np < 7 );

	       if ( pchk || dchk )
	       {
		  OTI_Nvekt(np, spol, dv, nvekt);

		  if (pchk)
		     mp = OTI_planp(np, spol, nvekt);

		  if (dchk && mp)
		     mp = OTI_pland(np, spol, dv, nvekt);
	       }

	       if (cchk && mp)
		  mp = OTI_planc(np, spol, dv);

	       if (mp)
	       {
		  for (i = 0;  i < GRD;  i++)
		     if (q = p->n.nxt[i])
		     {
			if (q->typ == B)
			{
			   if ((e = TMA_ele(ktab, p->Orig, &j)) && j == 1)
			      e->m = 0;	       /* Zelle ist behandelt ! */
			   else
			   {
			      printf("!! Fehler (OTS_join): Element nicht gefunden!\n");
			      exit();
			   }
			}

			OTS_del(q);  p->n.nxt[i] = NULL;
		     }

		  OTS_ekn(p,B);
	       }
	    }
         }
      }
   }
}

void
OTS_tcheck(p)		      /* Untersuche den Knoten p gemaess topologischer Unterteilungskriterien */
OTknoten *p;		      /* (E) zu untersuchender Knoten */
{
   short int i, j, k,
	     is, jf, jx, jy,
	     ks;
   OTgk kd, kdn, key[DIM], mp[4][DIM];
   OTgr rs, rf;
   OTknoten *pn, *q;
   OTpunkt  mw[4];
   TMelem *e;

   if (p->typ == G)
   {
      if (p->level >= otw.mlv1)
      {
	 kd = OTA_gkd(p);

         for (is = 0;  is < 12;  is++)	     /* 1. Kantentest */
         {
	    rs = OTrstr[is];

            if (p->n.nxt[ OTA_drodr(rs,0) ] || p->n.nxt[ OTA_drodr(rs,1) ])
            {
               OTI_dstr(p, kd, is);

	       if ((ks = SIGNUM(str.wa)) == SIGNUM(str.we))
               {
                  OTI_assgk(key, str.pa);  i = str.kv;
		  key[i] += str.pe[i];
		  key[i] >>= 1;

		  if ( !(e = TMA_ele(ktab, key, &j)) || j != 1)
		  {
		     printf("!! Fehler (OTS_tcheck/S): Element nicht gefunden!\n");
		     exit();
		  }

                  if (SIGNUM(e->w) != ks)
                     for (i = 0;   i < 3;  i++)
                        while ((pn = (i ? OTA_gsn(p,OTA_orodr(rs,i-1)) : OTA_gen(p,rs))) &&
			      !(pn->level == p->level && pn->typ == G))
                        {
			   kdn = OTA_gkd(pn);
                           OTS_div(pn, kdn);
			   pn->ksect = -1;
			   OTS_part(pn);
                        }
               }
            }
         }  /* Ende for is... */

         for (jf = 0;  jf < 6;  jf++)	     /* 2. Seitentest */
         {
	    rf = OTrfac[jf];

	    for (i = j = 0;  i < 4 && j == 0;  i++)
	       j |= (p->n.nxt[OTA_dror(rf,i)] != NULL);

	    if (j && (pn = OTA_gsn(p,rf)) &&
		    !(pn->level == p->level && pn->typ == G))
	    {
	       OTI_dfac(p, kd, jf);

	       for (j = 1, i = 0;  i < 3 && j;  i++)
		  j &= (SIGNUM(fac.w[i]) == SIGNUM(fac.w[i+1]));

               if (j)		 /* Vorzeichen der Eckpunkte sind gleich */
               {
                  OTI_assgk(key, fac.pkt[0]);

		  for (i = 1;   i < 4;  i++)
                     for (j = 0;   j < 3;  j++)
			key[j] += fac.pkt[i][j];

		  for (j = 0; j < 3;  j++)
		     key[j] >>= 2;	     /* Mittelpunkt der Facette */

		  i = (rf - XX) >> 1;	     /* Index der konstanten Koordinate */
		  jx = (i+1) % 3;
		  jy = (i+2) % 3;
		  k  = kd >> 1;

		  for (i = 0; i < 4;  i++)
		     OTI_assgk(mp[i], key);

		  mp[0][jx] -= k;
		  mp[1][jy] -= k;
		  mp[2][jx] += k;
		  mp[3][jy] += k;

		  for (i = 0; i < 4;  i++)
		  {
		     if ( !(e = TMA_ele(ktab, mp[i], &j)) || j != 1)
		     {
			printf("!! Fehler (OTS_tcheck/F1): Element nicht gefunden!\n");
			exit();
		     }

		     mw[i] = e->w;
		  }

		  for (j = 1, i = 0;  i < 3 && j;  i++)
		     j &= (SIGNUM(mw[i]) == SIGNUM(mw[i+1]));

		  if (j && SIGNUM(mw[0]) == SIGNUM(fac.w[0]))
		  {		 /* Vorzeichen der Eckpunkte und Kantenmittelpunkte sind gleich */
		     if ( !(e = TMA_ele(ktab, key, &j)) || j != 1)
		     {
			printf("!! Fehler (OTS_tcheck/F2): Element nicht gefunden!\n");
			exit();
		     }

		     if (SIGNUM(mw[0]) != SIGNUM(e->w))
                        do
                        {
			   kdn = OTA_gkd(pn);
                           OTS_div(pn, kdn);
			   pn->ksect = -1;
			   OTS_part(pn);
                        } while ((pn = OTA_gsn(p,rf)) &&
				!(pn->level == p->level && pn->typ == G));
		  }
               }
	    }
	 }  /* Ende for jf... */
      }

      for (i = 0;  i < GRD;  i++)
	 if ((q = p->n.nxt[i]) && q->typ == G)
	    OTS_tcheck(q);
   }
}

void
OTS_Vqueue(mod)		      /* Verwalte Knoten-Queue */
short mod;		      /* (E) Modus (1: initialisieren, 2: loeschen) */
{
   if (mod)
      FXkub = FXS_def( 30, sizeof(OTknoten *), 20);
   else
      FXS_loe(FXkub);

   Nqueue = 0;
}

short			      /* 0: ok, -1: Fehler */
OTS_push(p)		      /* Knotenverweis an Queue anhaengen */
OTknoten *p;		      /* (E) Verweis auf Knoten */
{
   if (FXS_ael(FXkub, &p) != FXkub)
      return ( -1 );

   Nqueue++;
   return ( 0 );
}

OTknoten *
OTS_pop()		      /* Knotenverweis von Queue holen */
{
   if (Nqueue > 0)
   {
      FLEX *z;
      OTknoten *p;

      z = FXA_ele(FXkub, 0, &p);
      z = FXS_eel(FXkub, 0);
      --Nqueue;
      return ( z == FXkub ? p : NULL );
   }

   return ( NULL );
}

short
OTS_Spol(p, mod, kd, pol)     /* Schnittpolygon bestimmen */
OTknoten *p;		      /* (E) Verweis auf Knoten (Typ B) */
short mod;		      /* (E) Modus (0: Test, 1: Polygonerzeugung) */
OTgk kd;		      /* (E) aktuelle Koordinatendifferenz */
OTpol *pol;		      /* (A) Schnittpolygon */
{
   short int i, ii, j, jj, k, kk, kf,
	     ia, ie, is,
	     ka, ke, ks,
	     ja, je, js, j1,
	     isanf,
	     ip, np,
	     ifac, jfac, ifc1,
	     spez = 0;
   OTknoten *q, *q1, *pn;
   OTknset  pnset[MAXKNSET];
   OTgk kdf, rc[GR2];
   OTtyp kpv, kpn;
   TMelem *e0;

   if ( kub.p != p )
      OTI_dkub(p, kd);

   for ( i = 0; i < 12; i++ )
   {
      ka = kub.str[i][0];
      ke = kub.str[i][1];

      if ( SIGNUM(kub.w[ka]) != SIGNUM(kub.w[ke]))
	 break;
   }
assert( i < 12 );

   is = ( kub.w[ka] > 0.0 ? i+1 : -(i+1));

   if (is < 0)
   {
      k = ka;  ka = ke;  ke = k;
   }

   if (kub.w[ke] == 0.0)
   {				 /* Spezialbehandlung bei Ecken-Durchgang */
      ja   = (is > 0 ? 0 : 1);
      jfac = kub.sfr[i][ja];
      j    = kub.skr[i][ja];
      ip = 0;

      for ( ii = 0; ii < 3; ii++ )
      {
	 j = OTI_nxtfk(jfac, j, &js, &kk, &ja, &je);

	 if (kub.w[ja] > 0.0)
	    ip++;

	 if ( SIGNUM(kub.w[ja]) != SIGNUM(kub.w[je]))
	    break;
      }
assert( ii < 3 );

      if (ip == 3)
      {
	 is = js;  i = kk;  ka = ja;  ke = je;
      }
   }

   np = 0;  isanf = i;  q1 = NULL;
   ja   = ( is > 0 ? 0 : 1 );
   ifac = kub.sfr[i][ja];
   k  =   kub.str[i][2];
   ks = j = kub.skr[i][ja];

   do
   {
      for ( ii = 0; ii < 3; ii++ )
      {
	 j = OTI_nxtfk(ifac, j, &is, &kk, &ia, &ie);

	 if ( SIGNUM(kub.w[ia]) != SIGNUM(kub.w[ie]))
	    break;
      }
assert( ii < 3 );

      q = NULL;   kf = k;

      if (mod)
      {
	 pn = OTA_gsn(p, OTrfac[ifac]);

	 switch (kpv = OTA_mark(pn))
	 {
	    case T:
	    case B:
	       if (!(e0 = TMA_ele(ktab,pn->Orig,&je)) || je != 1)
	       {
		  printf("!! Fehler (OTS_Spol): Element nicht gefunden!\n");
		  exit();
	       }

	       if (e0->m < 0)
	       {
		  OTS_push(pn);
		  e0->m = 0;
	       }

               if (kpv == T)
               {
                  ka = OTgkkub[ks];
                  ke = (ka < 3 ? ka + 1 : 0);
		  OTI_dfac(pn, kd, OTcfac[ifac]);
		  OTI_conTk(OTgkkub[ii], ka, ke, &np, pol);
               }

	       break;

	    case G:
	    {
	       OTgk k1[DIM];

	       ip = 0;  q = NULL;  jj = OTgkkub[ks];
	       OTI_rcomp(OTrfac[ifac], rc);
	       OTI_rckn(pn, rc, &ip, pnset);

	       if (kub.w[ke] == 0.0)
	       {
		  OTI_assgk(k1, kub.pkt[ke]);
	       }
	       else
	       {
		  OTI_assgk(k1, kub.pkt[ka]);
	       }

	       repeat
	       {
		  if ((q = OTI_skn(pnset, k1, kf, kub.fac[OTcfac[ifac]][jj])) == NULL)
		  {
		     printf("!! Fehler (OTS_Spol/1): Nachbarknoten nicht gefunden\n");
		     printf("   Knoten: %d,%d,%d;%d\n", p->Orig[0], p->Orig[1], p->Orig[2], kd);
		     break;
		  }

		  if (!(e0 = TMA_ele(ktab,q->Orig,&je)) || je != 1)
		  {
		     printf("!! Fehler (OTS_Spol): Element nicht gefunden!\n");
		     exit();
		  }

		  if (e0->m < 0)
		  {
		     OTS_push(q);
		     e0->m = 0;
		  }

		  ka = ja = jj;
		  ke = (ja < 3 ? ja + 1 : 0);
		  kdf = OTA_gkd(q);
		  OTI_dfac(q, kdf, OTcfac[ifac]);

		  if (spez)
		     ja = ( ja > 0 ? ja - 1 : 3 );

		  if (np == 0)
		  {
		     q1 = q;  j1 = ja;  ifc1 = ifac;
		  }

		  for ( ii = 0; ii < 3; ii++ )
		  {
		     ja = (ja < 3 ? ja + 1 : 0);
		     je = (ja < 3 ? ja + 1 : 0);

		     if (SIGNUM(fac.w[ja]) != SIGNUM(fac.w[je]) ||
			 fac.w[je] == 0.0 && fac.str[ja] == k)
			break;
		  } 

                  if (spez && ii < 3)
                  {
                     if (OTA_mark(q) == T)
			OTI_conTk(ii, ka, ke, &np, pol);

		     OTI_Sp(k1, fac.w[ja], fac.pkt[je], fac.w[je], kf, kdf, &pol[np++]);

		     for ( ii++; ii < 3; ii++ )
		     {
			ja = (ja < 3 ? ja + 1 : 0);
			je = (ja < 3 ? ja + 1 : 0);

			if (SIGNUM(fac.w[ja]) != SIGNUM(fac.w[je]) ||
			    fac.w[je] == 0.0 && fac.str[ja] == k)
			   break;
		     } 
                  }

		  if (spez = (ii == 0 && fac.w[ja] == 0.0))
		  {
		     kf = fac.str[ja];
		     OTI_assgk(k1, fac.pkt[je]);
		  }
		  else if (fac.w[je] == 0.0)
		  {
		     kf = fac.str[je];
		     OTI_assgk(k1, fac.pkt[je]);
		  }
		  else
		  {
		     kf = fac.str[ja];
		     OTI_assgk(k1, fac.pkt[ja]);
		  }

		  if ( ii < 3 && !spez )
		  {			  /* naechste Schnittkante gefunden */
                     if (OTA_mark(q) == T)
			OTI_conTk(ii, ka, ke, &np, pol);

		     OTI_Sp(k1, fac.w[ja], fac.pkt[je], fac.w[je], kf, kdf, &pol[np++]);
		  }

		  if (OTI_kident(kub.pkt[ia], k1, kf))
		     break;

		  if (kub.w[ie] == 0.0 && OTI_kident(kub.pkt[ie], k1, kf))
		     break;

		  jj = (spez ? ja : OTgkfac[ja]);
	       }
	    }
         }
      }

      i = kk;

      if (kub.w[ie] == 0.0)
      {				 /* Spezialbehandlung bei Ecken-Durchgang */
	 ja = (is > 0 ? 0 : 1);
	 jfac = kub.sfr[i][ja];
	 j    = kub.skr[i][ja];
	 ip = 0;

	 for ( ii = 0; ii < 3; ii++ )
	 {
	    j = OTI_nxtfk(jfac, j, &js, &kk, &ja, &je);

	    if (kub.w[ja] > 0.0)
	       ip++;

	    if ( SIGNUM(kub.w[ja]) != SIGNUM(kub.w[je]))
	       break;
	 }

	 if (ip == 3 && kk != isanf)   /* keine Beh. wenn ie letzter Punkt ist */
	 {
	    is = js;  i = kk;  ia = ja;   ifac = jfac;
assert( ie == je );
	 }
      }

      ka = ia;  ke = ie;
      ja   = ( is > 0 ? 0 : 1 );
      jfac = ifac;
      ifac = kub.sfr[i][ja];
      k    = kub.str[i][2];
      ks = j = kub.skr[i][ja];

      if (q == NULL)				   /* akt. Facette hat keine unterteilte Nachbarfacette */
      {
	 if ( mod && kub.w[ie] != 0.0 && kpv != G )
	 {					   /* Schnittpunkt mit groesstmoeglicher Aufloesung suchen */
	    pn = OTA_gsn(p, OTrfac[ifac]);
	    kpn = OTA_mark(pn);

	    if (kpn == G && i == isanf)
	       if (q1)	 /* Korrektur des Polygon-Endpunkts */
	       {
		  kdf = OTA_gkd(q1);
		  OTI_dfac(q1, kdf, OTcfac[ifc1]);
		  ia = j1;  ie = (ia < 3 ? ia+1 : 0);
		  kf = fac.str[ia];
		  OTI_Sp(fac.pkt[ia], fac.w[ia], fac.pkt[ie], fac.w[ie], kf, kdf, &pol[np++]);
		  continue;
	       }

	    if (kpn == G)
	    {
	       ip = 0;  kk = OTgkkub[ks];
	       OTI_rcomp(OTrfac[ifac], rc);
	       OTI_rckn(pn, rc, &ip, pnset);

	       if ((q = OTI_skn(pnset, kub.pkt[ia], k, kub.fac[OTcfac[ifac]][kk])) == NULL )
	       {
		  printf("!! Fehler (OTS_Spol/2): Nachbarknoten nicht gefunden\n");
		  printf("   Knoten: %d,%d,%d;%d\n", p->Orig[0], p->Orig[1], p->Orig[2], kd);
		  break;
	       }

	       kdf = OTA_gkd(q);
	       OTI_dfac(q, kdf, OTcfac[ifac]);
	       ja = kk;
	       je = ( ja < 3 ? ja+1: 0 );
	       kf = fac.str[ja];
	       OTI_Sp(fac.pkt[ja], fac.w[ja], fac.pkt[je], fac.w[je], kf, kdf, &pol[np++]);
	       continue;
	    }

	    if (kpn != G || i == isanf)
	    {
	       pn = OTA_gen(p, OTrstr[i]);

	       if (OTA_mark(pn) == G)
	       {
		  ip = 0;  kk = OTcstr[i] + 1;
		  OTI_rcomp(OTrstr[i], rc);
		  OTI_rckn(pn, rc, &ip, pnset);

		  if ((q = OTI_skn(pnset, kub.pkt[ia], k, kk)) == NULL)
		  {
		     printf("!! Fehler (OTS_Spol/3): Nachbarknoten nicht gefunden\n");
		     printf("   Knoten: %d,%d,%d;%d\n", p->Orig[0], p->Orig[1], p->Orig[2], kd);
		     break;
		  }

		  kdf = OTA_gkd(q);
		  OTI_dstr(q, kdf, OTcstr[i]);
		  OTI_Sp(str.pa, str.wa, str.pe, str.we, str.kv, kdf, &pol[np++]);
		  continue;
	       }
	    }
	 }

	 OTI_Sp(kub.pkt[ia], kub.w[ia], kub.pkt[ie], kub.w[ie], k, kd, &pol[np++]);
      }
   } while (i != isanf);

   return ( np );
}

short
OTS_Stetra(p, ifc, pol)	      /* Schnittpolygon bestimmen (fuer Tetraeder) */
OTknoten *p;		      /* (E) Verweis auf Knoten (Typ T) */
short *ifc;		      /* (E/A) Facettenindex */
OTpol *pol;		      /* (A) Schnittpolygon */
{
   short int i, ii, j, k, kk, itet,
	     ia, ie, is,
	     ka, ke, ks,
	     ja, je, js, jj,
	     isanf,
	     ip, np,
	     ifac, jfac;
   OTknoten *pn;
   OTtyp kpv;
   TMelem *e0;

   if ( *ifc == 0 && kub.p != p )
   {
      OTI_dkub(p, 2);
      OTI_assgk(tet.pkt[3], p->Orig);

      for ( i = 0; i < DIM; i++ )
	 tet.pkt[3][i] += 1;

      if (!(e0 = TMA_ele(ktab,tet.pkt[3], &k)) || k != 1)
      {
	 printf("!! Fehler (OTS_Stetra): Element nicht gefunden!\n");
	 exit();
      }

      tet.w[3] = e0->w;
   }

   repeat
   {
      if ((i = *ifc) >= 6)
         return ( 0 );

      for ( kk = itet = 0; itet < 4; itet++ )
      {
	 kk |= (p->n.tet[i][itet] == 'B');

	 if (p->n.tet[i][itet] == 'S')
	    break;
      }

      if (itet < 4)
	 break;
      else
	 (*ifc)++;
   }

   if (!kk)
   {
      pn = OTA_gsn(p, OTrfac[i]);

      if ((kpv = OTA_mark(pn)) == B || kpv == T)
      {
	 if (!(e0 = TMA_ele(ktab, pn->Orig, &k)) || k != 1)
	 {
	    printf("!! Fehler (OTS_Stetra): Element nicht gefunden!\n");
	    exit();
	 }

	 if (e0->m < 0)
	 {
	    OTS_push(pn);
	    e0->m = 0;
	 }
      }

      k  = kub.fac[i][0];
      ia = ( k > 0 ? kub.str[k-1][0] : kub.str[-k-1][1] );
      k  = kub.fac[i][2];
      ie = ( k > 0 ? kub.str[k-1][0] : kub.str[-k-1][1] );

      for ( j = 0; j < DIM; j++ )	  /* Facettenmittelpunkt bestimmen */
	 tet.pkt[2][j] = (kub.pkt[ia][j] + kub.pkt[ie][j]) >> 1;

      if (!(e0 = TMA_ele(ktab, tet.pkt[2], &j)) || j != 1)
      {
	 printf("!! Fehler (OTS_Stetra): Element nicht gefunden!\n");
	 exit();
      }

      tet.w[2] = e0->w;
   }

   k  = kub.fac[i][itet];

   if (k > 0)
   {
      ia = kub.str[k-1][0];
      ie = kub.str[k-1][1];
   }
   else
   {
      ia = kub.str[-k-1][1];
      ie = kub.str[-k-1][0];
   }
   
   OTI_assgk(tet.pkt[0], kub.pkt[ia]);   tet.w[0] = kub.w[ia];
   OTI_assgk(tet.pkt[1], kub.pkt[ie]);   tet.w[1] = kub.w[ie];

   for ( i = 0; i < 6; i++ )
   {
      ka = tet.str[i][0];
      ke = tet.str[i][1];

      if ( SIGNUM(tet.w[ka]) != SIGNUM(tet.w[ke]))
	 break;
   }
/* assert( i < 6 ); */
   if ( i > 5 )
   {
      np = 0;
      goto ende;
   }

   is = ( tet.w[ka] > 0.0 ? i+1 : -(i+1));

   if (is < 0)
   {
      k = ka;  ka = ke;  ke = k;
   }

   if (tet.w[ke] == 0.0)
   {				 /* Spezialbehandlung bei Ecken-Durchgang */
      ja   = (is > 0 ? 0 : 1);
      jfac = tet.sfr[i][ja];
      j    = tet.skr[i][ja];
      ip = 0;

      for ( ii = 0; ii < 2; ii++ )
      {
	 j = OTI_nTfk(jfac, j, &js, &kk, &ja, &je);

	 if (tet.w[ja] > 0.0)
	    ip++;

	 if ( SIGNUM(tet.w[ja]) != SIGNUM(tet.w[je]))
	    break;
      }
assert( ii < 2 );

      if (ip == 2)
      {
	 is = js;  i = kk;  ka = ja;  ke = je;
      }
   }

   np = 0;  isanf = i;
   ja   = ( is > 0 ? 0 : 1 );
   ifac = tet.sfr[i][ja];
   ks = j = tet.skr[i][ja];

   do
   {
      for ( ii = 0; ii < 2; ii++ )
      {
	 j = OTI_nTfk(ifac, j, &is, &kk, &ia, &ie);

	 if ( SIGNUM(tet.w[ia]) != SIGNUM(tet.w[ie]))
	    break;
      }
assert( ii < 2 );

      i = kk;

      if (tet.w[ie] == 0.0)
      {				 /* Spezialbehandlung bei Ecken-Durchgang */
	 ja = (is > 0 ? 0 : 1);
	 jfac = tet.sfr[i][ja];
	 j    = tet.skr[i][ja];
	 ip = 0;

	 for ( ii = 0; ii < 2; ii++ )
	 {
	    j = OTI_nTfk(jfac, j, &js, &kk, &ja, &je);

	    if (tet.w[ja] > 0.0)
	       ip++;

	    if ( SIGNUM(tet.w[ja]) != SIGNUM(tet.w[je]))
	       break;
	 }

	 if (ip == 2 && kk != isanf)   /* keine Beh. wenn ie letzter Punkt ist */
	 {
	    is = js;  i = kk;  ia = ja;   ifac = jfac;
assert( ie == je );
	 }
      }

      ka = ia;  ke = ie;
      ja   = ( is > 0 ? 0 : 1 );
      jfac = ifac;
      ifac = tet.sfr[i][ja];
      ks = j = tet.skr[i][ja];

      OTI_Spnorth(tet.pkt[ia], tet.w[ia], tet.pkt[ie], tet.w[ie], &pol[np++]);
   } while (i != isanf);

ende:
   p->n.tet[*ifc][itet] = 'B';

   return ( np );
}

short int		      /* max. Tiefe, um Aufloesung epsd1 bei Seitenlaenge sl zu erreichen */
OTA_maxT (sl, epsd1, epsd2)
OTpunkt sl,		      /* (E) Seitenlaenge */
        epsd1,		      /* (E) gewuenschte obere Aufloesungsschranke */
       *epsd2;		      /* (A) gewaehlt Aufloesung */
{
   short int k;
   OTpunkt d;

   for (d = sl, k = 2;  k < 16 && d > epsd1;  k++)
      d = sl / (1 << k);

   *epsd2 = d;

   return ( k - 1 );
}

OTknoten *		   /* Verweis auf GSN-Nachbarn zu p in Richtung r */
OTA_gsn (p, r)
OTknoten *p;		   /* (E) vorgegebener Knoten */
OTgr r;			   /* (E) Suchrichtung (orthogonal) */
{
   OTgr rc[GR2];

   OTI_rcomp(r, rc);
   return ( OTI_gsn(p, r, rc));
}

OTknoten *		   /* Verweis auf GEN-Nachbarn zu p in Richtung r */
OTA_gen (p, r)
OTknoten *p;		   /* (E) vorgegebener Knoten */
OTgr r;			   /* (E) Suchrichtung (ortho-diagonal) */
{
   OTgr rc[GR2];

   OTI_rcomp(r, rc);
   return ( OTI_gen(p, r, rc, XXX));
}

OTknoten *		   /* Verweis auf GCN-Nachbarn zu p in Richtung r */
OTA_gcn (p, r)
OTknoten *p;		   /* (E) vorgegebener Knoten */
OTgr r;			   /* (E) Suchrichtung (diagonal) */
{
   OTgr cc;

   cc = OTI_ccomp(r);
   return ( OTI_gcn(p, r, cc, XXX));
}

OTknoten *		   /* Verweis auf Zelle mit linker unterer Ecke Orig */
OTA_elem (Orig)
OTgk Orig[];		   /* (E) Linke untere Ecke */
{
   short int i;
   OTknoten *p;
   OTgk kd, mp[DIM];
   OTgr r;

   p = otw.start;
   kd = OTA_gkd(p);

   while (p->typ == G)
   {
      kd >>= 1;

      for ( i = 0; i < DIM; i++ )
	 mp[i] = p->Orig[i] + kd;

      r = OTI_class(mp, Orig);
      p = p->n.nxt[r];
   }

   return (p);
}

OTknoten *		   /* Verweis auf nicht behandelte Zelle */
OTA_szell (t)
TMtab *t;		   /* (E) Verweis auf Koordinatentabelle */
{
   int i, j, n;
   short int k;
   OTknoten *q;
   TMelem *p;

   for (q = NULL, i = -1; (p = TMA_nxt(t, i, &i)) && i >= 0;)
   {
      if (p->m < 0 )			    /* Element noch nicht behandelt */
      {
	 q = OTA_elem(p->k);

	 for ( n = 1, j = 0; j < DIM && n; j++ )
	    n &= (q->Orig[j] == p->k[j]);

	 if (!n)
	    if (!(p = TMA_ele(ktab, q->Orig, &k)) || k != 1)
	    {
	       printf("!! Fehler (OTA_Szell): Element nicht gefunden!\n");
	       exit();
	    }

	 if (p->m < 0)			    /* Knoten noch nicht behandelt */
	    break;
	 else
	    q = NULL;
      }
   }

   return (q);
}

#ifdef XXX	/* z.Zt. nicht benoetigt */

OTknoten *		   /* Verweis auf unbehandelten Nachbar-Knoten in Richtung r */
OTA_Mnachbar (p, r, m)
OTpunkt *p;		   /* (E) Verweis auf Ausgangsknoten */
OTgr r;			   /* (E) Suchrichtung */
short int m;		   /* (E) Markierung des gesuchten Nachbarn */
{
   OTknoten *q;
   short int i, k;
   OTgr cc, rc[GR2];
   TMelem *e0;

   if (OTA_dr(r))			 /* diagonale Suchrichtung */
   {
      if (q = OTA_gcn(p, r))
      {
	 cc = OTI_ccomp(r);

	 while (OTA_mark(q) == G)
	    q = q->n.nxt[cc];

	 e0 = TMA_ele(ktab, q->Orig, &k);

	 if (!(m < 0 ? e0->m < 0 : e0->m == m))
	    q = NULL;
      }

      for ( i = 0; i < 6 && q == NULL; i++ )
      {				    /* moeglichen Verlauf durch Gitterpunkt pruefen */
	 cc = OTderg[r][i];
	 q  = OTA_Mnachbar(p, cc, m);
      }
   }
   else
   {				       /* orthogonale oder ortho-diagonale Suchrichtung */
      if (q = (OTA_or(r) ? OTA_gsn(p, r) : OTA_gen(p, r)))
      {
	 if (OTA_mark(q) == G)
	 {
	    OTI_rcomp(r, rc);
	    q = OTI_Melem(q, m, rc);
	 }
	 else
	 {
	    e0 = TMA_ele(ktab, q->Orig, &k);

	    if (!(m < 0 ? e0->m < 0 : e0->m == m))
	       q = NULL;
	 }
      }

      if ( OTA_odr(r))		 /* moeglichen Verlauf durch Gitter-Kanten pruefen */
	 for ( i = 0; i < 2 && q == NULL; i++ )
	    q  = OTA_Mnachbar(p, OToderg[r-20][i], m);
   }

   return ( q );
}
#endif

void
OTA_nel(p, nkg, nke)		 /* Anzahl definierter Knoten ? */
OTknoten *p;			 /* (E) Knoten, ab dem gezaehlt werden soll */
int *nkg,			 /* (A) Anzahl der Knoten gesamt */
    *nke;			 /* (A) Anzahl der Endknoten */
{
   if (p->typ == G)
   {
      short int i;

      for (i = 0; i < GRD; i++)
	 if (p->n.nxt[i])
	    OTA_nel(p->n.nxt[i], nkg, nke);
   }
   else
      (*nke)++;

   (*nkg)++;
}

unsigned int
OTA_NSele()
{
   return ( Nqueue );
}

#ifdef XXX	/* z.Zt. nicht benoetigt */

PRIVATE OTpunkt *
OTI_Melem (p, m, rc)	      /* markiertes Element ? */
OTknoten *p;
short int m;
OTgr rc[];
{
   OTknoten *q, *pn;
   OTgr rc2[GR2];
   short int i, k;
   TMelem *e;

   if (OTA_mark(p) == G)
   {
      for ( i = 0; i < GR2 && rc[i] != XXX; i++ )
	 if (q = p->n.nxt[rc[i]])
	 {
	    if (pn = OTI_Melem(q, m, rc))
	       return ( pn );
	 }
   }
   else if ( OTA_mark(p) == B )
   {
      e = TMA_ele(ktab, p->Orig, &k);

      if (m < 0 ? e->m < 0 : e->m == m)
	 return ( p );
   }

   return ( NULL );
}
#endif

PRIVATE short int	   /* Klassifikationscode der Schnittsituation */
OTI_ksect (p, kd)	   /* = 0: kein Schnitt, > 0: Schnitt eindeutig, < 0: Schnitt nicht eindeutig */
OTknoten *p;		   /* (E) Verweis auf zu betrachtenden Knoten */
OTgk kd;		   /* (E) aktuelle Koordinatendifferenz */
{
   short int i, j, k, kk, ks, ksec,
	     is, ip[4], i1, i2,
	     np, nz, nn;
   OTgk  kd2, key[DIM];
   OTpunkt wm;
   TMelem *e;

   for ( np = nn = nz = i = 0; i < GRD; i++ )
      if ( kub.w[i] > 0.0 )
	 np++;
      else if ( kub.w[i] < 0.0 )
	 nn++;
      else
	 nz++;

   if ( np == 8 || nn == 8 ||
       (np == 7 || nn == 7) && nz == 1 ||
       (np == 6 || nn == 6) && nz == 2 )
   {
      ksec = ( schk && kd > 2 ? -1 : 0 );
   }
   else
   {
      nn += nz;
      ks = np*10 + nn;  k = 1;
      ksec = -1;		      /* ggf. zu falsifizieren */

      switch (ks)		      /* Fallunterscheidung auf Basis Normalform */
      {
	 case 71:
	 case 17:
	    ksec = ks;
	    break;

	 case 62:  k = -1;
	 case 26:
	    for ( i = 0; i < GRD; i++ )
	       if ( SIGNUM(kub.w[i]) == k)
		  break;

	    for ( j = 0; j < 3; j++ )
	    {
	       kk = kub.psr[i][j];

	       if ( SIGNUM(kub.w[kub.str[kk][0]]) ==
		    SIGNUM(kub.w[kub.str[kk][1]]))
	       {
		  ksec = ks;
		  break;
	       }
	    }
	    break;

	 case 53:  k = -1;
	 case 35:
	    for ( i = 0; i < 7; i++ )
	       if ( SIGNUM(kub.w[i]) == k)
		  break;

	    for ( j = OTirel3[i]; j < OTirel3[i+1]; j++ )
	       if ((kk = SIGNUM(kub.w[OTprel3[j][0]])) ==
		    SIGNUM(kub.w[OTprel3[j][1]]) && k == kk )
	       {
		  ksec = ks;
		  break;
	       }

	    break;

	 case 44:
	    for ( i = 0; i < 6 && ksec < 0; i++ )
	    {
	       k = SIGNUM(kub.w[i]);

	       for ( j = OTirel4[i]; j < OTirel4[i+1]; j++ )
		  if ((kk = SIGNUM(kub.w[OTprel4[j][0]])) ==
			    SIGNUM(kub.w[OTprel4[j][1]]) &&
		       k == kk &&
		       k == SIGNUM(kub.w[OTprel4[j][2]]))
		  {
		     ksec = ks;
		     break;
		  }
	    }
	    break;

	 default:
	    printf("Fehler (OTI_ksect): unzulaessiger Fall: %d\n", ks );
      }
   }

   if (schk && ksec > 0 && kd <= 4)
   {
      if (ksec != 44)			  /* Wuerfelmitte auf Konsistenz pruefen */
      {
         kd2 = kd >> 1;

         for (i = 0;  i < DIM;  i++)
            key[i] = p->Orig[i] + kd2;

	 if ((e = TMS_ins(ktab, key, &j)) == NULL || j < 0)
	 {
	    printf("!! Fehler (OTI_ksect): Speichermangel!\n");
	    exit();
	 }
	 else if (j == 0)
	    OTS_gkv(e);

         if (ksec < 44)
         {
            if (e->w >= 0.0)
	       ksec = -1;
         }
         else
         {
            if (e->w <= 0.0)
	       ksec = -1;
         }
      }

      for (i = 0;  i < 6 && ksec > 0;  i++)
      {
         for (np = nn = j = 0;  j < 4;  j++)
         {
            is = kub.fac[i][j];
	    ip[j] = ( is > 0 ? kub.str[is-1][0] : kub.str[-is-1][1] );

            if (kub.w[ip[j]] > 0.0)
               np++;
            else
               nn++;
         }

         if (np == 4 || nn == 4)
         {				  /* Mittelpunkt der Wuerfelseiten testen */
            for (j = 0;  j < DIM;  j++)
               key[j] = (kub.pkt[ip[0]][j] + kub.pkt[ip[2]][j]) >> 1;

            if ((e = TMS_ins(ktab, key, &j)) == NULL || j < 0)
            {
	       printf("!! Fehler (OTI_ksect): Speichermangel!\n");
	       exit();
            }
	    else if (j == 0)
	       OTS_gkv(e);

	    for (wm = kub.w[ip[3]], j = 0;  j < 3;  j++)
	       if ( kub.w[ip[j]] > wm )
		  wm = kub.w[ip[j]];

	    if (e->w > wm)
	       ksec = -1;
         }
      }

      for (i = 0;  i < 12 && ksec > 0;  i++)
      {				    /* Wuerfelkanten pruefen */
         i1 = kub.str[i][0];  i2 = kub.str[i][1];

         if ((k = SIGNUM(kub.w[i1])) == SIGNUM(kub.w[i2]))
         {
            for (j = 0;  j < DIM;  j++)
               key[j] = (kub.pkt[i1][j] + kub.pkt[i2][j]) >> 1;

	    if ((e = TMS_ins(ktab, key, &j)) == NULL || j < 0)
	    {
	       printf("!! Fehler (OTI_ksect): Speichermangel!\n");
	       exit();
	    }
	    else if (j == 0)
	       OTS_gkv(e);

            if (k == 1)
            {
	       wm = (kub.w[i1] < kub.w[i2] ? kub.w[i1] : kub.w[i2]);

	       if (e->w * 1.6 < wm)
		  ksec = -1;
            }
            else
            {
	       wm = (kub.w[i1] > kub.w[i2] ? kub.w[i1] : kub.w[i2]);

	       if (e->w * 1.6 > wm)
		  ksec = -1;
            }
         }
      }
   }  /* Ende schk */

   return ( ksec );
}

PRIVATE void
OTI_kTsect (p)		   /* Klassifikation der Tetraeder-Schnittsituation */
OTknoten *p;		   /* (E) Verweis auf zu betrachtenden Knoten */
{
   short int i, j, k,
	     ifc, it, i0, i2,
	     np, nz, nn;
   TMelem *e;
   OTgk key[DIM];

   for ( i = 0; i < DIM; i++ )	    /* Wuerfelmittelpunkt bestimmen */
      key[i] = p->Orig[i] + 1;

   if ((e = TMS_ins(ktab, key, &j)) == NULL || j < 0)
   {
      printf("!! Fehler (OTI_kTsect): Speichermangel!\n");
      exit();
   }
   else if (j == 0)
      OTS_gkv(e);

   tet.w[3] = e->w;

   for (ifc = 0;  ifc < 6;  ifc++)
   {
      k = kub.fac[ifc][0];
      i0 = (k > 0 ? kub.str[k-1][0] : kub.str[-k-1][1]);
      k = kub.fac[ifc][2];
      i2 = (k > 0 ? kub.str[k-1][0] : kub.str[-k-1][1]);

      for ( j = 0; j < DIM; j++ )	  /* Facettenmittelpunkt bestimmen */
	 key[j] = (kub.pkt[i0][j] + kub.pkt[i2][j]) >> 1;

      if ((e = TMS_ins(ktab, key, &j)) == NULL || j < 0)
      {
	 printf("!! Fehler (OTI_kTsect): Speichermangel!\n");
	 exit();
      }
      else if (j == 0)
	 OTS_gkv(e);

      tet.w[2] = e->w;
      tet.w[0] = kub.w[i0];

      for ( it = 0; it < 4; tet.w[0] = tet.w[1], it++ )
      {
	 k  = kub.fac[ifc][it];
	 i2 = ( k > 0 ? kub.str[k-1][1] : kub.str[-k-1][0] );
	 tet.w[1] = kub.w[i2];

	 for ( np = nn = nz = i = 0; i < 4; i++ )
	    if ( tet.w[i] > 0.0 )
	       np++;
	    else if ( tet.w[i] < 0.0 )
	       nn++;
	    else
	       nz++;

	 if ( np == 4 || nn == 4 ||
	     (np == 3 || nn == 3) && nz == 1 ||
	     (np == 2 || nn == 2) && nz == 2 ||
	      nz == 3 && nn == 1 )
	    p->n.tet[ifc][it] = 'L';
	 else
	    p->n.tet[ifc][it] = 'S';
      }
   }
}

PRIVATE void
OTI_conTk(ik, ia, ie, np, pol)	    /* Verbinden der Schnittpunkte mit Tetraederkanten */
short ik,			    /* (E) Index der Zielkante */
      ia, ie,			    /* (E) Anfangs- und Endindex der Punkte der Ausgangskante */
      *np;			    /* (E/A) Anzahl der Punkte des Schnittpolygons */
OTpol pol[];			    /* (E/A) Schnittpolygon */
{
   short i, in;
   TMelem *e;
   OTpunkt  mw;
   OTgk	mp[DIM];

   for ( i = 0; i < DIM; i++ )
      mp[i] = (fac.pkt[0][i] + fac.pkt[2][i]) >> 1;

   if (!(e = TMA_ele(ktab, mp, &i)) || i != 1)
   {
      printf("!! Fehler (OTI_conTk): Element nicht gefunden!\n");
      exit();
   }

   mw = e->w;

   if (ik == 1)
   {
      if (fabs(mw) < epsd3)
	 OTI_Spnorth(mp, 0.0, mp, 0.0, &pol[(*np)++]);
      else
      {
	 if (SIGNUM(fac.w[ie]) != SIGNUM(mw))
	 {
	    OTI_Spnorth(fac.pkt[ie], fac.w[ie], mp, mw, &pol[(*np)++]);
	    in = (ie < 3 ? ie + 1 : 0);

	    if (SIGNUM(fac.w[in]) != SIGNUM(mw))
	       OTI_Spnorth(fac.pkt[in], fac.w[in], mp, mw, &pol[(*np)++]);
	    else
	       printf("!! Warnung (OTI_conTk): Nur eine Tetraederkante geschnitten\n");
	 }
	 else if (SIGNUM(fac.w[ia]) != SIGNUM(mw))
	 {
	    OTI_Spnorth(fac.pkt[ia], fac.w[ia], mp, mw, &pol[(*np)++]);
	    in = (ia > 0 ? ia - 1 : 3);

	    if (SIGNUM(fac.w[in]) != SIGNUM(mw))
	       OTI_Spnorth(fac.pkt[in], fac.w[in], mp, mw, &pol[(*np)++]);
	    else
	       printf("!! Warnung (OTI_conTk): Nur eine Tetraederkante geschnitten\n");
	 }
	 else
	    printf("!! Warnung (OTI_conTk): Keine Tetraederkante geschnitten\n");
      }
   }
   else
   {
      in = (ik ? ia : ie);

      if (SIGNUM(fac.w[in]) != SIGNUM(mw))
         OTI_Spnorth(fac.pkt[in], fac.w[in], mp, mw, &pol[(*np)++]);
      else if (fabs(mw) < epsd3)
	 OTI_Spnorth(mp, 0.0, mp, 0.0, &pol[(*np)++]);
      else
      {
	 in = (!ik ? ia : ie);

	 for ( i = 0; i < 3; i++ )
	 {
	    if (SIGNUM(fac.w[in]) != SIGNUM(mw))
	       OTI_Spnorth(fac.pkt[in], fac.w[in], mp, mw, &pol[(*np)++]);
	    
	    if (ik)
	       in = (in < 3 ? in + 1 : 0);
	    else
	       in = (in > 0 ? in - 1 : 3);
	 }
      }
   }
}

PRIVATE OTgr		/* Richtung von p2 bezueglich p1 */
OTI_class (p1, p2)
OTgk p1[],		/* (E) 1. Punkt */
     p2[];		/* (E) 2. Punkt */
{
   short xn, yn, zn;

   xn = (p2[0] < p1[0]);
   yn = (p2[1] < p1[1]);
   zn = (p2[2] < p1[2]);
   return (xn ? (yn ? (zn ? UVW : UVZ) : (zn ? UYW : UYZ))
	      : (yn ? (zn ? XVW : XVZ) : (zn ? XYW : XYZ)));
}

void
OTI_rcomp (r, rc)	      /* Komplementaere Diagonal-Richtungen zu einer
				 orthogonal- bzw. ortho-diagonal-Richtung ? */
OTgr r,			      /* (E) Richtung */
     rc[];		      /* (A) Komplementaer-Richtungen */
{
   if (OTA_odr(r))
      rc[2] = rc[3] = XXX;

   switch (r)
   {
      case XX:  rc[0] = UYZ;  rc[1] = UYW;  rc[2] = UVZ;  rc[3] = UVW;  break;
      case UU:  rc[0] = XYZ;  rc[1] = XYW;  rc[2] = XVZ;  rc[3] = XVW;  break;
      case YY:  rc[0] = XVZ;  rc[1] = UVZ;  rc[2] = XVW;  rc[3] = UVW;  break;
      case VV:  rc[0] = XYZ;  rc[1] = UYZ;  rc[2] = XYW;  rc[3] = UYW;  break;
      case ZZ:  rc[0] = XYW;  rc[1] = UYW;  rc[2] = XVW;  rc[3] = UVW;  break;
      case WW:  rc[0] = XYZ;  rc[1] = UYZ;  rc[2] = XVZ;  rc[3] = UVZ;  break;
      case XY:  rc[0] = UVZ;  rc[1] = UVW;  break;
      case UY:  rc[0] = XVZ;  rc[1] = XVW;  break;
      case UV:  rc[0] = XYZ;  rc[1] = XYW;  break;
      case XV:  rc[0] = UYZ;  rc[1] = UYW;  break;
      case XZ:  rc[0] = UYW;  rc[1] = UVW;  break;
      case XW:  rc[0] = UYZ;  rc[1] = UVZ;  break;
      case UW:  rc[0] = XYZ;  rc[1] = XVZ;  break;
      case UZ:  rc[0] = XYW;  rc[1] = XVW;  break;
      case YZ:  rc[0] = XVW;  rc[1] = UVW;  break;
      case VZ:  rc[0] = XYW;  rc[1] = UYW;  break;
      case VW:  rc[0] = XYZ;  rc[1] = UYZ;  break;
      case YW:  rc[0] = XVZ;  rc[1] = UVZ;  break;
      default:  rc[0] = XXX;
   }
}

PRIVATE OTgr
OTI_rref(r, rr)		      /* gespiegelte Richtung bezueglich einer     */
			      /* orthogonal- oder ortho-diagonal-Richtung  */
OTgr r, rr;		      /* (E) diagonal-Richtung und Relativrichtung */
{
assert( OTA_dr(r) );

   switch (rr)
   {
      case XX:
      case UU:
         return ( OTrref[0][r] );

      case YY:
      case VV:
         return ( OTrref[1][r] );

      case ZZ:
      case WW:
         return ( OTrref[2][r] );

      case XY:
      case UV:
      case XV:
      case UY:
	 return ( OTrref[3][r] );

      case XZ:
      case UW:
      case XW:
      case UZ:
	 return ( OTrref[4][r] );

      case YZ:
      case VW:
      case VZ:
      case YW:
	 return ( OTrref[5][r] );

      default:
         return (XXX);
   }
}

PRIVATE OTgr		      /* Richtung, in der ein Knoten von seinem Vorgaenger abgeht */
OTI_rvorg (p)
OTknoten *p;		      /* (E) Verweis auf zu betrachtenden Knoten */
{
   short int i;
   OTknoten *q;

   if (q = p->vorg)
      for (i = 0;  i < GRD;  i++)
	 if (p == q->n.nxt[i])
	    return ( i );

   return ( XXX );	      /* darf nie vorkommen ! */
}

PRIVATE OTgr		      /* Gemeinsame Orthogonal- bzw. ortho-diagonal-Richtung */
OTI_rconst (r1, r2)	      /* zu zwei diagonal- bzw. ortho-diagonal-Richtungen */
OTgr r1, r2;		      /* (E) Richtungen */
{
   if (r1 > r2)
   {
      OTgr rh;

      rh = r2; r2 = r1; r1 = rh;
   }
   else if ( r1 == r2 )
      return (XXX);

   switch (r1)
   {
      case XYZ:
	 switch (r2)
	 {
	    case XVZ:   return ( XZ );
	    case UYZ:   return ( YZ );
	    case XYW:   return ( XY );
	    case XV:
	    case XW:
	    case XVW:   return ( XX );
	    case UY:
	    case YW:
	    case UYW:   return ( YY );
	    case UZ:
	    case VZ:
	    case UVZ:   return ( ZZ );
	    case XY:
	    case XZ:
	    case YZ:	return ( r2 );
	    default:    return ( XXX );
	 }
      case UYZ:
	 switch (r2)
	 {
	    case UYW:   return ( UY );
	    case UVZ:   return ( UZ );
	    case UV:
	    case UW:
	    case UVW:   return ( UU );
	    case YW:
	    case XY:
	    case XYW:   return ( YY );
	    case XZ:
	    case VZ:
	    case XVZ:   return ( ZZ );
	    case UY:
	    case UZ:
	    case YZ:	return ( r2 );
	    default:    return ( XXX );
	 }
      case XYW:
	 switch (r2)
	 {
	    case UYW:   return ( YW );
	    case XVW:   return ( XW );
	    case XV:
	    case XZ:
	    case XVZ:   return ( XX );
	    case UY:
	    case YZ:    return ( YY );
	    case UW:
	    case VW:
	    case UVW:   return ( WW );
	    case XY:
	    case XW:
	    case YW:	return ( r2 );
	    default:    return ( XXX );
	 }
      case UYW:
	 switch (r2)
	 {
	    case UVW:   return ( UW );
	    case UV:
	    case UZ:
	    case UVZ:   return ( UU );
	    case YZ:
	    case XY:    return ( YY );
	    case XW:
	    case VW:
	    case XVW:   return ( WW );
	    case UY:
	    case UW:
	    case YW:	return ( r2 );
	    default:    return ( XXX );
	 }
      case XVZ:
	 switch (r2)
	 {
	    case UVZ:   return ( VZ );
	    case XVW:   return ( XV );
	    case XY:
	    case XW:    return ( XX );
	    case UV:
	    case VW:
	    case UVW:   return ( VV );
	    case UZ:
	    case YZ:    return ( ZZ );
	    case XV:
	    case XZ:
	    case VZ:	return ( r2 );
	    default:    return ( XXX );
	 }
      case UVZ:
	 switch (r2)
	 {
	    case UVW:   return ( UV );
	    case UY:
	    case UW:    return ( UU );
	    case XV:
	    case VW:
	    case XVW:   return ( VV );
	    case XZ:
	    case YZ:    return ( ZZ );
	    case UV:
	    case UZ:
	    case VZ:	return ( r2 );
	    default:    return ( XXX );
	 }
      case XVW:
	 switch (r2)
	 {
	    case UVW:   return ( VW );
	    case XY:
	    case XZ:	return ( XX );
	    case UV:
	    case VZ:	return ( VV );
	    case UW:
	    case YW:	return ( WW );
	    case XV:
	    case XW:
	    case VW:	return ( r2 );
	    default:    return ( XXX );
	 }
      case UVW:
	 switch (r2)
	 {
	    case UY:
	    case UZ:	return ( UU );
	    case XV:
	    case VZ:	return ( VV );
	    case XW:
	    case YW:	return ( WW );
	    case UV:
	    case UW:
	    case VW:	return ( r2 );
	    default:    return ( XXX );
	 }

      case XY:
	 switch (r2)
	 {
	    case UY:
	    case YZ:
	    case YW:    return ( YY );
	    case XV:
	    case XZ:
	    case XW:    return ( XX );
	    default:    return ( XXX );
	 }
      case UY:
	 switch (r2)
	 {
	    case UV:
	    case UZ:
	    case UW:    return ( UU );
	    case YZ:
	    case YW:    return ( YY );
	    default:    return ( XXX );
	 }
      case UV:
	 switch (r2)
	 {
	    case XV:
	    case VZ:
	    case VW:    return ( VV );
	    case UZ:
	    case UW:    return ( UU );
	    default:    return ( XXX );
	 }
      case XV:
	 switch (r2)
	 {
	    case XZ:
	    case XW:    return ( XX );
	    case VZ:
	    case VW:    return ( VV );
	    default:    return ( XXX );
	 }
      case XZ:
	 switch (r2)
	 {
	    case XW:    return ( XX );
	    case UZ:
	    case YZ:
	    case VZ:    return ( ZZ );
	    default:    return ( XXX );
	 }
      case XW:
	 switch (r2)
	 {
	    case UW:
	    case VW:
	    case YW:    return ( WW );
	    default:    return ( XXX );
	 }
      case UW:
	 switch (r2)
	 {
	    case UZ:    return ( UU );
	    case VW:
	    case YW:    return ( WW );
	    default:    return ( XXX );
	 }
      case UZ:
	 switch (r2)
	 {
	    case YZ:
	    case VZ:    return ( ZZ );
	    default:    return ( XXX );
	 }
      case YZ:
	 switch (r2)
	 {
	    case VZ:    return ( ZZ );
	    case YW:    return ( YY );
	    default:    return ( XXX );
	 }
      case VW:
	 switch (r2)
	 {
	    case YW:    return ( WW );
	    default:    return ( XXX );
	 }

      default:   return ( XXX );
   }
}

PRIVATE OTknoten *		 /* Verweis auf GEN-Nachbarn */
OTI_gen (p, r, rc, rvv)
OTknoten *p;
OTgr r, 
     rc[],
     rvv;
{
   OTknoten *q;
   OTgr rv, rr, rg, rrk;
   short i, k;

   if (q = p->vorg)
   {
      rv = OTI_rvorg(p);
assert(OTA_dr(rv));

      for ( k = 1, i = 0; i < GR2 && rc[i] != XXX; i++ )
	 k &= (rv != rc[i]);

      rr = OTI_rref(rv, r);

      if ( rvv == XXX )
	 rvv = r;

      if (k)
	 if (rv != rvv)
	 {
	    rg = OTI_rconst(rv, rvv);

	    if (OTA_or(rg))
	       rrk = rg;
	    else if ( r != rg)
	    {
	       rrk = OTI_rconst(r,rg);
assert( OTA_or(rrk));
	    }
	    else
	       rrk = 0;

	    if (rrk)
	       q = OTA_gsn(q, rrk);
	    else
	       q = OTI_gen(q, r, rc, rv);
	 }
	 else
	    q = OTI_gen(q, r, rc, rv);
   }

   if (q)
      if (q->n.nxt[rr])
	 return (q->n.nxt[rr]);

   return ( q );
}

PRIVATE OTknoten *		 /* Verweis auf GSN-Nachbarn */
OTI_gsn (p, r, rc)
OTknoten *p;
OTgr r, 
     rc[];
{
   OTknoten *q;
   OTgr rv, rr;
   short i, k;

   if (q = p->vorg)
   {
      rv = OTI_rvorg(p);
assert(OTA_dr(rv));

      for ( k = 1, i = 0; i < GR2 && rc[i] != XXX; i++ )
	 k &= (rv != rc[i]);

      rr = OTI_rref(rv, r);

      if (k)
	 q = OTI_gsn(q, r, rc);
   }

   if (q)
      if (q->n.nxt[rr])
	 return (q->n.nxt[rr]);

   return ( q );
}

PRIVATE OTknoten *		 /* Verweis auf GCN-Nachbarn */
OTI_gcn (p, c, cc, cvv)
OTknoten *p;
OTgr c, 
     cc, cvv;
{
   OTknoten *q;
   OTgr cv;

   if (q = p->vorg)
   {
      cv = OTI_rvorg(p);
assert(OTA_dr(cv));

      if (cv == c)
	 q = OTI_gcn(q, c, cc, cv);
      else if (cv != cc)
      {
	 OTgr r;

	 if ( cvv == -1 )
	    cvv = c;

	 r = OTI_rconst(cv, cvv);
assert(r != XXX);
	 q = OTA_gsn(q, r);
      }
   }

   if (q)
   {
      cc = OTI_ccomp(cv);

      if (q->n.nxt[cc])
	 return (q->n.nxt[cc]);
   }

   return ( q );
}

short int		      /* Rueckmeldung: 0 = ok, < 0 = Fehler */
OTI_upd (t, Orig, kd)	      /* Innere Gitterpunkte in Tabelle eintragen */
TMtab *t;		      /* (E) Verweis auf aktuelle Tabelle */
OTgk  Orig[],		      /* (E) Ursprung in Gitterkoordinaten */
      kd;		      /* (E) Koordinaten-Differenz */
{
   short int j, ni1, ni2, ni3;
   OTgk  key[DIM];
   TMelem *e;

   for ( key[2] = Orig[2], ni3 = 0; ni3 < 3; key[2] += kd, ni3++ )
      for ( key[1] = Orig[1], ni2 = 0; ni2 < 3; key[1] += kd, ni2++ )
	 for ( key[0] = Orig[0], ni1 = 0; ni1 < 3; key[0] += kd, ni1++ )
	 {
	    if (!(e = TMS_ins(ktab, key, &j)) || j < 0)
	       return(j);
	    else if (j == 0)
	       OTS_gkv(e);
	 }

   return ( 0 );
}

PRIVATE void
OTI_vzw (t, mod, Orig, kd, km, mp, mn)  /* Pruefen der Eckpunkte einer quadratischen Zelle auf Vorzeichenwechsel */
TMtab *t;				/* (E) Tablle */
short int mod;				/* (E) Modus */
OTgk  Orig[],				/* (E) Linke untere Ecke */
      kd,				/* (E) aktuelle Koordinaten-Differenz */
      km;				/* (E) minimale Koordinaten-Differenz */
short int *mp, *mn;			/* (E/A) Signal: Alle Eckpunkt-Markierungen positiv/negativ */
{
   short int i, 
	     j;
   int	 k;
   OTgk  key[DIM];
   TMelem *q;
/* global:
   kt[], ir
*/

   for ( i = 0; i < DIM; i++ )
      key[kt[ir][i]] = Orig[kt[ir][i]];

/*
   Definition der Arbeits-Modi:
      Bei Modus 2 bedeutet '.n' die Rekursionstiefe.

           3----0----0----0----3                              
           |    |    |    |    |                              
          2.2--2.2--2.2--2.2---1                              
           |    |    |    |    |                             
          2.1--2.2--2.1--2.2---1                             
           |    |    |    |    |                             
          2.2--2.2--2.2--2.2---1                             
           |    |    |    |    |                             
           3---2.2--2.1--2.2---3                             
                                                             
                +-km-+                                       
           +---------kd--------+                             
*/
   switch (mod)
   {
      case 3:			    /* Vorzeichenwechsel an den Ecken der akt. Zelle ? */
	 for (i = 0;  i < 4 && ( *mp || *mn );  i++)
	 {
	    switch (i)
	    {
	       case 1:
		  key[kt[ir][0]] += kd;
		  break;
	       case 0:
		  key[kt[ir][1]] += kd;
		  break;
	       case 2:
		  key[kt[ir][0]] = Orig[kt[ir][0]];
		  key[kt[ir][1]] = Orig[kt[ir][1]];
		  break;
	       case 3:
		  key[kt[ir][0]] += kd;
	    }

	    if ( !(q = TMA_ele(t, key, &j)) || j != 1)
	       goto Error;

	    if ( *mp ) *mp &= ( q->w > .0 );
	    if ( *mn ) *mn &= ( q->w <= .0 );
	 }
	 break;

      case 2:			/* Vzw an den Orig-Ecken der Quadranten 0,1,3 ? */
	 if ((k = (kd >> 1)) >= km)
	 {
	    OTgk  k2[4][DIM];

	    for (i = 0;  i < 4 && ( *mp || *mn );  i++)
	    {
	       switch (i)
	       {
		  case 1:
		     key[kt[ir][0]] += k;
		     break;
		  case 0:
		     key[kt[ir][1]] += k;
		     break;
		  case 2:
		     key[kt[ir][0]] = Orig[kt[ir][0]];
		     key[kt[ir][1]] = Orig[kt[ir][1]];
		     break;
		  case 3:
		     key[kt[ir][0]] += k;
	       }

	       if (i != 2)
	       {
		  if ( !(q = TMA_ele(t, key, &j)) || j != 1)
		     goto Error;

		  if ( *mp ) *mp &= ( q->w > .0 );
		  if ( *mn ) *mn &= ( q->w <= .0 );
	       }

	       k2[i][kt[ir][0]] = key[kt[ir][0]];
	       k2[i][kt[ir][1]] = key[kt[ir][1]];
	       k2[i][kt[ir][2]] = key[kt[ir][2]];
	    }

	    if ((*mp || *mn) && k > km)		/* innere Punkte des aktuellen Quadranten pruefen */
	       for (i = 0; i < 4 && (*mp || *mn); i++)
		  OTI_vzw(t, mod, k2[i], k, km, mp, mn);
	 }
	 break;

      case 0:
      case 1:			/* Vzw am inneren oberen/rechten Rand der akt. Zelle ? */
	 key[kt[ir][0]] += kd;  key[kt[ir][1]] += kd;

	 for (i = km;  i < kd && ( *mp || *mn );  i += km)
	 {
	    key[kt[ir][mod]] -= km;

	    if ( !(q = TMA_ele(t, key, &j)) || j != 1)
	       goto Error;

	    if ( *mp ) *mp &= ( q->w > .0 );
	    if ( *mn ) *mn &= ( q->w <= .0 );
	 }
   }
   return;

Error:
   fprintf( stderr, "Fehler (OTI_vzw) Element nicht gefunden!\n" );
   OTI_Exit();
}

PRIVATE void
OTI_vw3 (t, mod, Orig, kd, km, mp, mn)  /* Pruefen der Eckpunkte einer Wuerfel-Zelle auf Vorzeichenwechsel */
TMtab *t;				/* (E) Tablle */
short int mod;				/* (E) Modus */
OTgk  Orig[],				/* (E) Ursprungs-Ecke */
      kd,				/* (E) aktuelle Koordinaten-Differenz */
      km;				/* (E) minimale Koordinaten-Differenz */
short int *mp, *mn;			/* (E/A) Signal: Alle Eckpunkt-Markierungen positiv/negativ */
{
   short int i, 
	     j;
   int	 k;
   OTgk  key[DIM];
   TMelem *q;
/* global:
   ir
*/

   for ( i = 0; i < DIM; i++ )
      key[i] = Orig[i];

   switch ( mod )
   {
      case ECKEN:
	 ir = 0;
	 OTI_vzw(t, 3, key, kd, km, mp, mn );

	 if ( *mp || *mn )
	 {
	    key[2] += kd;
	    OTI_vzw(t, 3, key, kd, km, mp, mn );
	 }
	 break;

      case SEITEN:
	 for ( ir = 0; ir < DIM && (*mp || *mn); ir++ )
	 {
	    key[(ir+2)%3] += kd;
	    OTI_vzw(t, 2, key, kd, km, mp, mn );

	    if ( *mp || *mn )
	       OTI_vzw(t, 1, key, kd, km, mp, mn );

	    key[(ir+2)%3] -= kd;
	 }

	 ir = 0;
	 break;

      case INNEN:	      /* VZW an inneren Gitterpunkten ? */
	 if ((k = (kd >> 1)) >= km)
	 {
	    OTgk  k2[GRD][DIM];

	    for (i = 0;  i < GRD && ( *mp || *mn );  i++)
	    {
	       switch (i)
	       {
		  case 0:
		  case 5:
		  case 7:
		     key[0] += k;	break;
		  case 1:
		  case 3:
		     key[1] += k;	break;
		  case 2:
		  case 6:
		     key[0] = Orig[0];
		     key[1] = Orig[1];
		     break;
		  case 4:
		     key[2] += k;
	       }

	       if (i != 2)
	       {
		  if ( !(q = TMA_ele(t, key, &j)) || j != 1)
		     goto Error;

		  if ( *mp ) *mp &= ( q->w > .0 );
		  if ( *mn ) *mn &= ( q->w <= .0 );
	       }

	       OTI_assgk(k2[i], key);
	    }

	    if ((*mp || *mn) && k > km)		/* innere Punkte des aktuellen Oktanten pruefen */
	       for (i = 0; i < GRD && (*mp || *mn); i++)
		  OTI_vw3(t, mod, k2[i], k, km, mp, mn);
	 }
   }
   return;

Error:
   fprintf( stderr, "Fehler (OTI_vw3) Element nicht gefunden!\n" );
   exit();
}

PRIVATE void
OTI_dkub(p, kd)		/* Belegen der Ecken des zu p gehoerenden Kubus */
OTknoten *p;		/* (E) Verweis auf Knoten */
OTgk kd;		/* (E) aktuelle Koordinatendifferenz */
{
   short int i, k;
   OTgk key[DIM];
   TMelem *pp;

   kub.p = p;
   OTI_assgk(key, p->Orig);   OTI_assgk( kub.pkt[3], key );
   key[0] += kd;	      OTI_assgk( kub.pkt[0], key );
   key[1] += kd;	      OTI_assgk( kub.pkt[1], key );
   key[0] -= kd;	      OTI_assgk( kub.pkt[2], key );
   key[2] += kd;	      OTI_assgk( kub.pkt[6], key );
   key[1] -= kd;	      OTI_assgk( kub.pkt[7], key );
   key[0] += kd;	      OTI_assgk( kub.pkt[4], key );
   key[1] += kd;	      OTI_assgk( kub.pkt[5], key );

   for ( i = 0; i < 8; i++ )
   {
      if (!(pp = TMA_ele( ktab, kub.pkt[i], &k )) || k != 1)
      {
	 printf("!! Fehler (OTI_dkub): Element nicht gefunden!\n");
	 exit();
      }

      kub.w[i] = pp->w;
   }
}

PRIVATE void
OTI_gkoord(p, kd, ip, pkt)	 /* Gitterkoordinaten bzgl. Punktindex bestimmen */
OTknoten *p;			 /* (E) Verweis auf Bezugsknoten */
short ip;			 /* (E) Index des gesuchten Punktes */
OTgk kd,			 /* (E) aktuelle Koordinatendifferenz */
     pkt[];			 /* (A) Koordinaten des Punktes */
{
   OTI_assgk(pkt, p->Orig);

   if (ip > 3)
      pkt[2] += kd;

   switch (ip)
   {
      case 1:
      case 5:
         pkt[1] += kd;
      case 0:
      case 4:
         pkt[0] += kd;
         break;
      case 2:
      case 6:
         pkt[1] += kd;
   }
}

PRIVATE void
OTI_dstr(p,kd,ks)	/* Belegen der Strecken-Struktur mit den Werten der Kubus-Strecke ks */
OTknoten *p;		/* (E) Verweis auf Bezugsknoten */
OTgk kd;		/* (E) aktuelle Koordinatendifferenz */
short ks;		/* (E) Index der Kubus-Strecke */
{
   short int i, k, ia, ie;
   TMelem *pp;

   ia = kub.str[ks][0];
   ie = kub.str[ks][1];
   str.kv = kub.str[ks][2];
   OTI_gkoord(p, kd, ia, str.pa);

   if (!(pp = TMA_ele( ktab, str.pa, &k )) || k != 1)
   {
      printf("!! Fehler (OTI_dstr): Element nicht gefunden!\n");
      exit();
   }

   str.wa = pp->w;
   OTI_gkoord(p, kd, ie, str.pe);

   if (!(pp = TMA_ele( ktab, str.pe, &k )) || k != 1)
   {
      printf("!! Fehler (OTI_dstr): Element nicht gefunden!\n");
      exit();
   }

   str.we = pp->w;
}

PRIVATE void
OTI_dfac(p,kd,kf)	/* Belegen der Facetten-Struktur mit den Werten der Kubus-Facette kf */
OTknoten *p;		/* (E) Verweis auf Bezugsknoten */
OTgk kd;		/* (E) aktuelle Koordinatendifferenz */
short kf;		/* (E) Index der Kubus-Facette */
{
   short int i, k, ip, is;
   TMelem *pp;

   for ( i = 0; i < 4; i++ )
   {
      is = kub.fac[kf][i];
      ip = ( is > 0 ? kub.str[is-1][0] : kub.str[-is-1][1] );
      OTI_gkoord(p, kd, ip, fac.pkt[i]);

      if (!(pp = TMA_ele( ktab, fac.pkt[i], &k )) || k != 1)
      {
	 printf("!! Fehler (OTI_dfac): Element nicht gefunden!\n");
	 exit();
      }

      fac.w[i] = pp->w;
      fac.str[i] = kub.str[abs((int) is)-1][2];
   }
}

PRIVATE short
OTI_kident(p1, p2, kv)		 /* Kante durch p1 mit Kante durch p2 in Richtung kv identisch? */
OTgk  p1[], p2[];		 /* (E) Koordinaten der Testpunkte */
short kv;			 /* (E) Index der variierenden Achse */
{
   short i, h;

   for ( h = 1, i = 0; i < DIM && h; i++ )
      if ( i != kv )
	 h &= ( p1[i] == p2[i] );

   return (h);
}

PRIVATE void
OTI_rckn(p, rc, ip, pset)	 /* Endknotenmenge (B) in den Richtungen {rc} bzgl. des Knotens p bestimmen */
OTknoten *p;			 /* (E) Verweis auf Bezugsknoten */
OTknset  pset[];		 /* (E/A) Menge der gefundenen Endknoten */
OTgk rc[];			 /* (E) Richtungen der gesuchten Endknoten */
short *ip;			 /* (E/A) Anzahl gefundener Endknoten in pset */
{
   short ir;
   OTknoten *pn;

   for ( ir = 0; ir < GR2 && rc[ir] != XXX; ir++ )
   {
      pn = p->n.nxt[rc[ir]];

      switch (OTA_mark(pn))
      {
         case B:
         case T:
	    if ((*ip)+1 < MAXKNSET)
	    {
	       pset[*ip].done = 0;
	       pset[(*ip)++].kn = pn;
	    }
	    else
	    {
	       printf("Fehler (OTI_rckn): Knotenmenge zu gross!\n");
	       pset[*ip].kn = NULL;
	       return;
	    }
            break;

         case G:
            OTI_rckn(pn, rc, ip, pset);
      }
   }

   pset[*ip].kn = NULL;
}

PRIVATE OTknoten *
OTI_skn(pnset, pa, kv, ks)	       /* Bestimmen des Nachbarknotens aus {pnset}, der mit p den Punkt sp gemein hat */
OTknset  pnset[];		       /* (E) Menge der Nachbarknoten */
OTgk pa[];			       /* (E) Koordinaten eines Punktes auf der gemeinsamen Schnittkante */
short kv,			       /* (E) Index der variierenden Achse */
      ks;			       /* (E) Nummer der Gegenkante */
{
   short i, k, ia, ie;
   OTgk  p1[DIM], p2[DIM], kd;
   OTknoten *pn;
   TMelem *e1, *e2;

   if (ks > 0)
   {
      i = ks - 1;  ia = 0;  ie = 1;
   }
   else
   {
      i = -ks - 1;  ia = 1;  ie = 0;
   }

   ia = kub.str[i][ia];
   ie = kub.str[i][ie];

   for ( i = 0; pnset[i].kn; i++ )
      if (!pnset[i].done)
      {
	 pn = pnset[i].kn;
	 kd = OTA_gkd(pn);
	 OTI_gkoord(pn, kd, ia, p1);

         if (OTI_kident(pa, p1, kv))
         {
	    OTI_gkoord(pn, kd, ie, p2);

	    if (OTI_kident(pa, p2, kv))
	    {
	       if (!(e1 = TMA_ele(ktab, p1, &k)) || k != 1)
	       {
		  printf("!! Fehler (OTI_skn): Element nicht gefunden!\n");
		  exit();
	       }

	       if (!(e2 = TMA_ele(ktab, p2, &k)) || k != 1)
	       {
		  printf("!! Fehler (OTI_skn): Element nicht gefunden!\n");
		  exit();
	       }

	       if (SIGNUM(e1->w) != SIGNUM(e2->w) ||
		   e1->w != 0.0 && e2->w == 0.0 ||
		   e2->w != 0.0 && e1->w == 0.0)
	       {
		  pnset[i].done = 1;
		  return (pn);
	       }
	    }
         }
      }

   return ( NULL );
}

PRIVATE OTpunkt
OTI_mnorm(p1, p2)
OTpunkt  p1[], p2[];
{
   short    i;
   OTpunkt  h, mn;

   for ( mn = 0.0, i = 0; i < DIM; i++)
      if ((h = fabs(p1[i]-p2[i])) > mn)
	 mn = h;

   return (mn);
}

PRIVATE void
OTI_Sp( k1, w1, k2, w2, k, kd, pol )   /* Berechnen eines Schnittpunkts */
OTgk  k1[],			       /* (E) Koordinaten des 1. Punkts */
      k2[],			       /* (E) Koordinaten des 2. Punkts */
      kd;			       /* (E) aktuelle Koordinatendifferenz */
short k;			       /* (E) Index der variierenden Achse */
OTpunkt w1,			       /* (E) Wert des 1. Punkts */
        w2;			       /* (E) Wert des 2. Punkts */
OTpol	*pol;			       /* (A) Schnittpunkt */
{
   short int i, j;
   OTpunkt pd, w;
   TMelpk *e;

   TMS_spk(k1, k2, pol->sk);

   if ((e = TMA_ele(ptab, pol->sk, &i)) && i == 1)
      OTI_assgk(pol->sp, e->pkt)
   else
   {
      if (w1 == 0.0)
	 OTI_asspk(pol->sp, k1)
      else if (w2 == 0.0)
	 OTI_asspk(pol->sp, k2)
      else
      {
	 pd = k2[k] - k1[k];

	 for ( i = 0; i < DIM; i++ )
	    if ( i != k )
	       pol->sp[i] = otw.Orig[i] + k1[i] * epsd2;

	 if (!Iter1)
	 {
	    w = w1 / ( w1 - w2 );

	    if ( w > 1.0 )
	       w = 1.0;

	    pol->sp[k] = otw.Orig[k] + ( k1[k] + w * pd ) * epsd2;
	 }
	 else
	 {
	    OTpunkt  wp, wv, spk, p1k, p2k, dvor, dakt, epsL;

	    p1k = otw.Orig[k] + k1[k] * epsd2;
	    p2k = otw.Orig[k] + k2[k] * epsd2;
	    dvor = fabs(pd * epsd2);
	    spk  = p1k;	   wv = w2;

	    if ((epsL = dvor * 0.05) < 0.0001)
	       epsL = 0.0001;

	    for (j = 0;  j <= Iter1;  j++)
	    {
#ifdef BINSEARCH
	       pol->sp[k] = ( p1k + p2k) * 0.5;
#else
	       w = w1 / (w1 - w2);
	       pol->sp[k] = p1k + w * (p2k - p1k);
#endif
	       dakt  = fabs(spk - pol->sp[k]);

	       if (fabs(wv) < epsd3 && dakt < epsL)
		  break;

	       spk = pol->sp[k];
	       wp = F_val(pol->sp);

	       if ( SigEQ(wp, w1))
	       {
		  p1k = spk;   w1 = wp;   
#ifndef BINSEARCH

		  if (SigEQ(wp, wv))
		     w2 *= 0.5;
#endif 
	       }
	       else
	       {
		  p2k = spk;   w2 = wp;
#ifndef BINSEARCH

		  if (SigEQ(wp, wv))
		     w1 *= 0.5;
#endif 
	       }

	       wv = wp;
	    }
	 }  /* Ende if (Iter1) ...  */
      }  /* Ende else (w1 != 0 && w2 != 0 ) ... */

      if ((e = TMS_ins(ptab, pol->sk, &i)) && i == 0)
      {
	 OTI_assgk(e->pkt, pol->sp);

         if (nv_mod)
	    if (e->nv = malloc( DIM * sizeof(TMval)))
	       OTI_Ngrad(e->pkt, (OTpunkt) (kd * epsd2 * 0.01), e->nv);
	    else
	    {
	       printf("!! Fehler (OTI_Sp): Speichermangel !\n");
	       exit();
	    }
      }
      else
	 printf ("!! Fehler (OTI_Sp): Rueckmeldung von TMS_ins: %d\n", i);
   }  /* Ende else TMA_ele... */
}

PRIVATE void
OTI_Spnorth( k1, w1, k2, w2, pol )  /* Berechnen eines Schnittpunkts (nicht orthogonal)*/
OTgk  k1[],			    /* (E) Koordinaten des 1. Punkts */
      k2[];			    /* (E) Koordinaten des 2. Punkts */
OTpunkt w1,			    /* (E) Wert des 1. Punkts */
        w2;			    /* (E) Wert des 2. Punkts */
OTpol	*pol;			    /* (A) Schnittpunkt */
{
   short int i, j;
   OTpunkt w;
   TMelpk *e;

   TMS_spk(k1, k2, pol->sk);

   if ((e = TMA_ele(ptab, pol->sk, &i)) && i == 1)
      OTI_assgk(pol->sp, e->pkt)
   else
   {
      if (w1 == 0.0)
	 OTI_asspk(pol->sp, k1)
      else if (w2 == 0.0)
	 OTI_asspk(pol->sp, k2)
      else
      {
	 if (!Iter2)
	 {
	    w = w1 / ( w1 - w2 );

	    if ( w > 1.0 )
	       w = 1.0;

	    for ( i = 0; i < DIM; i++ )
	       pol->sp[i] = otw.Orig[i] + (k1[i] + w * (k2[i] - k1[i])) * epsd2;
	 }
	 else
	 {
	    OTpunkt  wp, wv, spv[DIM], p1[DIM], p2[DIM], dvor, dakt, epsL;

	    OTI_asspk(p1, k1);
	    OTI_asspk(p2, k2);
	    OTI_assgk(spv, p1);
	    wv = w2;
	    dvor = OTI_mnorm(p1, p2);

	    if ((epsL = dvor * 0.05) < 0.0001)
	       epsL = 0.0001;

	    for (j = 0;  j <= Iter2;  j++)
	    {
#ifdef BINSEARCH
	       for ( i = 0; i < DIM; i++ )
		  pol->sp[i] = (p1[i] + p2[i]) * 0.5;
#else
	       w = w1 / (w1 - w2);

	       for ( i = 0; i < DIM; i++ )
		  pol->sp[i] = p1[i] + w * (p2[i] - p1[i]);
#endif

	       dakt = OTI_mnorm(pol->sp, spv);

	       if (fabs(wv) < epsd3 && dakt < epsL)
		  break;

	       OTI_assgk(spv, pol->sp);
	       wp = F_val(pol->sp);

	       if (SigEQ(wp,w1))
	       {
		  OTI_assgk(p1, pol->sp);   w1 = wp;
#ifndef BINSEARCH

		  if (SigEQ(wp, wv))
		     w2 *= 0.5;
#endif
	       }
	       else
	       {
		  OTI_assgk(p2, pol->sp);   w2 = wp;
#ifndef BINSEARCH

		  if (SigEQ(wp, wv))
		     w1 *= 0.5;
#endif
	       }

	       wv = wp;
	    }
	 }  /* Ende if (Iter2) ...  */
      }  /* Ende else (w1 != 0 && w2 != 0 ) ... */

      if ((e = TMS_ins(ptab, pol->sk, &i)) && i == 0)
      {
	 OTI_assgk(e->pkt, pol->sp);

         if (nv_mod)
	    if (e->nv = malloc( DIM * sizeof(TMval)))
	       OTI_Ngrad(e->pkt, (OTpunkt) (epsd2 * 0.01), e->nv);
	    else
	    {
	       printf("!! Fehler (OTI_Spnorth): Speichermangel !\n");
	       exit();
	    }
      }
      else
	 printf ("!! Fehler (OTI_Spnorth): Rueckmeldung von TMS_ins: %d\n", i);
   }  /* Ende else TMA_ele... */
}

PRIVATE void
OTI_Ngrad(fp, dv, nv)	      /* Normierter Gradient (approximiert) im Punkt fp ? */
OTpunkt	 fp[],		      /* (E) Fusspunkt des Gradienten */
	 dv,		      /* (E) aktuelle absolute Koordinatendifferenz */
	 nv[];		      /* (A) normierter Gradientenvektor */
{
   short int i;
   OTpunkt d, f0, f[DIM], p[DIM];

   OTI_assgk(p, fp);
   f0 = F_val(p);
   p[0] += dv;		      f[0] = F_val(p);
   p[0] -= dv;	p[1] += dv;   f[1] = F_val(p);
   p[1] -= dv;	p[2] += dv;   f[2] = F_val(p);

   for ( d = 0.0, i = 0; i < DIM; i++ )
   {
      nv[i] = (f[i] - f0) / dv;
      d += nv[i] * nv[i];
   }

   if ((d = sqrt(d)) > 0.0 )
      for ( i = 0; i < DIM; i++ )
	 nv[i] /= d;
}

PRIVATE OTpunkt
OTI_NAdist(fp, dv)	      /* Nichtalgebraischer Abstand des Punktes fp von der Flaeche? */
OTpunkt	 fp[],		      /* (E) Punkt */
	 dv;		      /* (E) aktuelle absolute Koordinatendifferenz */
{
   short int i;
   OTpunkt d, nvi, f0, f[DIM], p[DIM];

   OTI_assgk(p, fp);
   f0 = F_val(p);
   p[0] += dv;		      f[0] = F_val(p);
   p[0] -= dv;	p[1] += dv;   f[1] = F_val(p);
   p[1] -= dv;	p[2] += dv;   f[2] = F_val(p);

   for ( d = 0.0, i = 0; i < DIM; i++ )
   {
      nvi = (f[i] - f0) / dv;
      d += nvi * nvi;
   }

   if ((d = sqrt(d)) > 0.0 )
      return ( fabs(f0) / d );
   else
      return ( fabs(f0));
}

PRIVATE void
OTI_Nvekt(np, pol, dv, nv)	     /* Normalvektor eines Polygons ? */
short np;
OTpol	pol[];			     /* (E) Schnittpolygon */
OTpunkt dv,			     /* (E) aktuelle absolute Koordinatendifferenz */
	nv[];			     /* (A) Normalvektor des Polygons */
{
   short int i, j;
   OTpunkt mp[DIM];

   OTI_assgk(mp, pol[0].sp);

   for ( i = 1; i < np; i++ )
      for ( j = 0; j < DIM; j++ )
	 mp[j] += pol[i].sp[j];

   for ( j = 0; j < DIM; j++ )
      mp[j] /= np;

   OTI_Ngrad(mp, dv, nv);
}

PRIVATE double
OTI_dvekt(p1, p2, nv)		    /* Kosinus des Winkels zwischen einem Verbindungsvektor und einem Normalvektor ? */
OTpunkt p1[], p2[],		    /* (E) Ortsvektoren */
	nv[];			    /* (E) Normalvektor */
{
   short int i;
   OTpunkt v[DIM], d, h;

   for ( d = h = 0.0, i = 0; i < DIM; i++ )
   {
      v[i] = p2[i] - p1[i];
      h += v[i] * v[i];
   }

   if ((h = sqrt(h)) > 0.0 )
      for ( i = 0; i < DIM; i++ )
      {
	 v[i] /= h;
	 d += v[i] * nv[i];
      }

   return ( d );
}

PRIVATE short
OTI_planp(np, pol, nv)		     /* Polygon plan bzgl. epsdp ? */
short np;			     /* (E) Anzahl der Polygonpunkte */
OTpol	 pol[];			     /* (E) Schnittpolygon */
OTpunkt  nv[];			     /* (E) Normalvektor des Polygons */
{
   short int i, j;
   OTpunkt dv;

   for ( i = 0; i < np; i++ )
   {
      j = ( i ? i-1 : np-1 );

      if ( !((dv = fabs(OTI_dvekt(pol[j].sp, pol[i].sp, nv))) < epsdp ))
	 return ( 0 );
   }

   return ( 1 );
}

PRIVATE short
OTI_pland(np, pol, dv, nv)	     /* Polygon plan bzgl epsdd ? */
short np;			     /* (E) Anzahl der Polygonpunkte */
OTpol	pol[];			     /* (E) Schnittpolygon */
OTpunkt dv,			     /* (E) aktuelle absolute Koordinatendifferenz */
	nv[];			     /* (E) Normalvektor des Polygons */
{
   short int i, j;
   OTpunkt d, dd;
   TMelpk  *e;

   for ( d = 1.0, i = 0; i < np; i++ )
   {
      if ((e = TMA_ele(ptab, pol[i].sk, &j)) && j == 1)
      {
	 for ( dd = 0.0, j = 0; j < DIM; j++ )
	    dd += e->nv[j] * nv[j];

	 if ( dd < d )
	 {
	    if ( !( 1.0 - dd < epsdd ))
	       return ( 0 );

	    d = dd;
	 }
      }
      else
	 printf("!! Warnung (OTI_pland): Element nicht gefunden!\n");
   }

   return ( 1 );
}

PRIVATE short
OTI_planc(np, pol, dv)		     /* Polygon plan bzgl. epsdc ? */
short np;			     /* (E) Anzahl der Polygonpunkte */
OTpol	pol[];			     /* (E) Schnittpolygon */
OTpunkt dv;			     /* (E) aktuelle absolute Koordinatendifferenz */
{
   short int i, j, k;
   OTpunkt dc, mp[DIM];

   for ( i = 0; i < np - 1; i++ )
      for ( j = i+1; j < np; j++ )
      {
	 for ( k = 0; k < DIM; k++ )
	    mp[k] = (pol[i].sp[k] + pol[j].sp[k]) * 0.5;

	 if ( !((dc = fabs(OTI_NAdist(mp, dv))) < epsdc ))
	    return ( 0 );
      }

   return ( 1 );
}

PRIVATE short			       /* relativer Kanten-Index der naechsten Facettenkante (Hexaeder) */
OTI_nxtfk(ifac, j, is, kk, ia, ie)     /* Endpunktindizes der naechsten Facettenkante ? */
short ifac,			       /* (E) Index der Facette */
      j,			       /* (E) relativer Kanten-Index der Facette */
      *is,			       /* (A) Nummer und Orientierung der Kante */
      *kk,			       /* (A) Index der Kante */
      *ia, *ie;			       /* (A) Indizes von Anfangs- und Endpunkt der Kante */
{
   short ja, je;

   j = (j < 3 ? j+1 : 0);

   if ((*is = kub.fac[ifac][j]) > 0)
   {
      *kk = (*is) - 1;  ja = 0;  je = 1;
   }
   else
   {
      *kk = -(*is) - 1;  ja = 1;  je = 0;
   }

   *ia = kub.str[*kk][ja];
   *ie = kub.str[*kk][je];
   return ( j );
}

PRIVATE short			       /* relativer Kanten-Index der naechsten Facettenkante (Tetraeder) */
OTI_nTfk(ifac, j, is, kk, ia, ie)     /* Endpunktindizes der naechsten Facettenkante ? */
short ifac,			       /* (E) Index der Facette */
      j,			       /* (E) relativer Kanten-Index der Facette */
      *is,			       /* (A) Nummer und Orientierung der Kante */
      *kk,			       /* (A) Index der Kante */
      *ia, *ie;			       /* (A) Indizes von Anfangs- und Endpunkt der Kante */
{
   short ja, je;

   j = (j < 2 ? j+1 : 0);

   if ((*is = tet.fac[ifac][j]) > 0)
   {
      *kk = (*is) - 1;  ja = 0;  je = 1;
   }
   else
   {
      *kk = -(*is) - 1;  ja = 1;  je = 0;
   }

   *ia = tet.str[*kk][ja];
   *ie = tet.str[*kk][je];
   return ( j );
}

/* fr Debugging ... */

#ifdef vms
#include unixio
#include unixlib
#endif

short
OTS_Tixe()
{
   short k;
   int	 n, nn,
	 fd;
   char	 *txt;
   TMkey    key[K2DIM];
   TMelem   *p;
   TMelpk   *q;
   unsigned sp, sv, sk1, sk2;

   if (txt = getenv("OT_LTAB"))
   {
      sk1 = KDIM * sizeof(TMkey);
      sk2 = K2DIM * sizeof(TMkey);
      sp  = KDIM * sizeof(TMval);
      sv  = sizeof(TMval);

      if ((fd = open("ktab.dmp", 0, 0)) >= 0)
      {
         repeat
         {
            if (read(fd,key,sk1) < sk1)
	       break;

            if ((p = TMS_ins(ktab,key,&k)) && k == 0)
            {
	       if (read(fd,&p->w,sv)  < sv )
                  printf("!! Fehler (OTS_Tixe): Lesefehler in ktab!\n");
            }
            else
            {
	       printf("!! Fehler (OTS_Tixe): Fehler beim Tabelleneintrag in ktab!\n");
               break;
            }
         }

	 close(fd);
      }
      else
	 printf("!! Fehler (OTS_Tixe): ktab.dmp nicht zugreifbar!\n");

      if ((fd = open("ptab.dmp", 0, 0)) >= 0)
      {
         repeat
         {
            if (read(fd,key,sk2) < sk2)
	       break;

            if ((q = (TMelpk *) TMS_ins(ptab,key,&k)) && k == 0)
            {
	       if (read(fd,q->pkt,sp)  < sp )
                  printf("!! Fehler (OTS_Tixe): Lesefehler in ptab!\n");
            }
            else
            {
	       printf("!! Fehler (OTS_Tixe): Fehler beim Tabelleneintrag in ptab!\n");
               break;
            }
         }

	 close(fd);
      }

      return (0);
   }
   return (1);
}

PRIVATE void
OTI_Exit()
{
   int	 n, nn,
	 fd;
   char	 *txt;
   TMelem   *p;
   TMelpk   *q;
   unsigned sp, sv, sk1, sk2;

   if (txt = getenv("OT_DUMP"))
   {
      sk1 = KDIM * sizeof(TMkey);
      sk2 = K2DIM * sizeof(TMkey);
      sp  = KDIM * sizeof(TMval);
      sv  = sizeof(TMval);

      if ((fd = creat("ktab.dmp", 0)) >= 0)
      {
         for (n = -1;  (p = TMA_nxt(ktab, n, &nn)) && nn >= 0;  n = nn)
         {
            if (write(fd,p->k,sk1) < sk1)    goto Error;
            if (write(fd,&p->w,sv)  < sv )    goto Error;
         }

	 close(fd);
      }
      else
	 printf("!! Fehler (OTI_Exit): ktab.dmp nicht erzeugbar!\n");

      if ((fd = creat("ptab.dmp", 0)) >= 0)
      {
         for (n = -1;  (q = (TMelpk *) TMA_nxt(ptab, n, &nn)) && nn >= 0;  n = nn)
         {
            if (write(fd,q->k,sk2)  < sk2)    goto Error;
            if (write(fd,q->pkt,sp) < sp )    goto Error;
	 /* hier muesste noch der Normalvektor ausgegeben werden !
	    (wird z.Zt. noch nicht benoetigt) */
         }

	 close(fd);
      }
      else
	 printf("!! Fehler (OTI_Exit): ptab.dmp nicht erzeugbar!\n");
   }

   exit();

   Error:
      printf("!! Fehler (OTI_Exit): Fehler beim Schreiben!\n");
   exit();
}
