/*
**
**  MACRO DEFINITIONS
**
**  OTS_mark(p, m)	Markieren des Knotens p mit Markierung m
**  OTA_mark(p)		Markierung des Knotens p (-1, falls p undefiniert)
**  OTA_gkd(p)		Gitter-Weite zum Knoten p
**  OTA_level(p)	Level des Knotens p
**  OTA_or(r)		r ist Orthogonal-Richtung
**  OTA_dr(r)		r ist Diagonal-Richtung
**  OTA_odr(r)		r ist Ortho-Diagonal-Richtung
**  OTA_crd(r)		komplementaere Richtung zur Richtung r
**  OTA_orodr(r,i)	Orthogonalrichtungen zu einer Ortho-Diagonal-Richtung
**  OTA_drodr(r,i)	Diagonalrichtungen zu einer Ortho-Diagonal-Richtung
**  OTA_dror(r,i)	Diagonalrichtungen zu einer Orthogonal-Richtung
*/

#define OTS_mark(p, m)	if (p) p->typ=m
#define OTA_mark(p)	(p ? p->typ : -1)
#define OTA_gkd(p)	(p ? 1 << (otw.mlev - p->level) : -1)
#define OTA_level(p)	(p ? p->level : -1)
#define OTA_or(r)	(10 <= r && r <= 15 ? 1:0)
#define OTA_dr(r)	(0  <= r && r < GRD ? 1:0)
#define OTA_odr(r)	(20 <= r && r <= 31 ? 1:0)
#define OTA_crd(r)	(OTA_or(r) ? OTscomp[r-XX] : (OTA_dr(r) ? OTccomp[r] : -1))
#define OTA_orodr(r,i)	((OTA_odr(r) && i < 2 && i >= 0) ? OToderg[r-XY][i] : (OTgr) XXX)
#define OTA_drodr(r,i)	((OTA_odr(r) && i < 2 && i >= 0) ? OTdrodr[r-XY][i] : (OTgr) XXX)
#define OTA_dror(r,i)	((OTA_or(r)  && i < 4 && i >= 0) ? OTdror[r-XX][i] : (OTgr) XXX)

#define DIM	3		 /* aktuelle Dimension	      */
#define GRD	8		 /* aktueller Grad des Baumes */
#define GR2	4		 /* halber Baum-Grad          */

/* Richtungscodierung: */

#define XYZ	0	/* Diagonal-Richtungen */
#define UYZ	1
#define XYW	2
#define UYW	3
#define XVZ	4
#define UVZ	5
#define XVW	6
#define UVW	7
#define XX	10	/* Orthogonal-Richtungen */
#define UU	11
#define YY	12
#define VV	13
#define ZZ	14
#define WW	15
#define XY	20	/* Ortho-Diagonal-Richtungen */
#define UY	21
#define UV	22
#define XV	23
#define XZ	24
#define XW	25
#define UW	26
#define UZ	27
#define YZ	28
#define VZ	29
#define VW	30
#define YW	31
#define XXX	-1

#ifndef NULL
#define NULL (char *) 0
#endif

#define repeat for(;;)
#define PRIVATE static

typedef double	     OTpunkt;
typedef short int    OTgk;	 /* Gitter-Koordinaten */
typedef short int    OTgr;	 /* Gitter-Richtungen  */

typedef enum
{
   W, B, G, T
} OTtyp;


#define W	0	/* white */
#define B	1	/* black */
#define G	2	/* grey */
#define T	3	/* Tetraeder */

typedef struct otknoten
{
   struct otknoten  *vorg;	/* Verweis auf Vorgaenger-Knoten (NIL, falls Wurzel ) */

   union
   {
      struct otknoten *nxt[GRD];/* Verweis auf Nachfolger-Knoten */
      char            tet[6][4];/* Markierung fuer Tetraeder-Schnitt (L: leer, S: Schnitt, B: bearbeitet) */
   } n;

   OTgk		    Orig[DIM];	/* Ursprung der Zelle in Gitter-Koordinaten */
   OTtyp	    typ;	/* Knoten-Typ */
   short int	    level,	/* aktueller OT-Level */
		    ksect;	/* aktuelle Schnittkennung */
} OTknoten;			/* Octree-Knoten */

typedef struct {
   OTknoten	*start;		/* Verweis auf Start-Knoten */
   OTpunkt	Orig[DIM],	/* Ursprung in absoluten Koordinaten */
		sl,		/* Seitenlaenge des Arbeitsbereichs */
		epsd;		/* max. Aufloesung */
   short int	mlv1,		/* max. OT-Tiefe fuer Detail-Erkennung */
	        mlvr,		/* min. OT-Tiefe fuer Reduktion */
		mlev;		/* max. OT-Tiefe fuer Kurven-Darstellung */
   TMkey	kd1max;		/* max. Koordinaten-Differenz auf Stufe 1 */
} OTwurzel;

typedef struct {
   OTpunkt  sp[DIM];		/* Polygon-Schnittpunkt */
   TMkey    sk[K2DIM];		/* Schluessel der zugehoerigen Wuerfelkante */
} OTpol;
